#ifndef LINT
static char vcid[] = "$Id: header.c,v 1.14 1994/10/20 22:11:04 gropp Exp $";
#endif

#include <ctype.h>
#include "tools.h"
#include "doc/doc.h"

/* 
   This is designed to work with comments in C programs.  
   It uses the standardized documentation to issue dummy routine
   definitions to allow the creation of a Fortran to C library.

   This version of "header.c" is a modification of the file taken 
   from "~gropp/tools.n/c2fort" on 10/7/93.  Modifications have been 
   introduced so that elements of type void * in the files "nlfunc_v.h", 
   and "nlspmat_v.h", and "nlsles_v.h" are translated as pointers to 
   structures in the Fortran version (instead of the default, which is
   no translation).  Note that pointers to void functions retain the 
   usual translation.  

   An additional flag (flag2) is used in the calling sequence of 
   ProcessArgDefs() to indicate the files which require the modified 
   translation.   Also, an additional element, void_function, is added 
   to the structure ARG_LIST to distinguish pointers to void functions 
   from pointers to void structures.  
 */
extern char GetSubClass();

static int NoFortMsgs = 1;
/* NoFortWarnings turns off messages about things not being available in
   Fortran */
static int NoFortWarnings = 1;

/* This says to convert char **a to int*a, and cast to (char **)*a */
static int MultipleIndirectAreInts    = 1;
static int MultipleIndirectsAreNative = 0;

/* Keep the file name to simplify finding files containing problems */
static char *CurrentFilename = 0;

/* Whether to replace pointers with indices to a mapping of pointers */
static int MapPointers = 0;
static char ptrprefix[32] = "__";

/* If this is set to true, "void *" are translated as pointers to structures */
static int TranslateVoidStar = 0;

/* If this is set to true, generate an "ifdef" block for the name of the
   Fortran interface to handle the case of no underscore and all caps
 */
static int IfdefFortranName = 0;

/* If true, add a last integer argument to int functions and return its
   value in the last parameter */
static int useFerr = 0;

/* Enable the MPI definitions */
static int isMPI = 0;

/* If 1, output header immediately; otherwise, defer until declarations
   processed */
static int OutputImmed = 1;

/* If 1, generate ANSI style headers instead */
static int AnsiHeader = 0;

/* If 0, do not generate the ifndef DEBUG_ALL wrapper */
static int AddDebugAll = 1;

/*D
  bfort - program to extract short definitions for a Fortran to C interface

  Input:
. filenames - Names the files from which lint definitions are to be extracted
. -nomsgs   - Do not generate messages for routines that can not be converted
              to Fortran.
. -nofort   - Generate messages for all routines/macros without a Fortran
              counterpart.
. -dir name - Directory for output file
. -I name   - file that contains common includes
. -mapptr   - translate pointers to integer indices
. -ptrprefix - prefix for names of functions to convert to/from pointers
              (default is __).
. -ferr     - Fortran versions return the value of the routine as the last
              argument (an integer).  This is used in MPI and is a not 
	      uncommon approach for handling error returns.
. -mpi      - Handle MPI datatypes (some things are pointers by definition)
. -mnative  - Multiple indirects are native datatypes (no coercion)
. -voidisptr - Consider "void *" as a pointer to a structure.
. -ansiheader - Generate ANSI-C style headers instead of Fortran interfaces
  This will be useful for createing ANSI prototypes   without ANSI-fying the 
  code.  These use a trick to provide both ANSI and non-ANSI prototypes.
  The declarations are wrapped in "ANSI_ARGS", the definition of which 
  should be
$  #ifdef ANSI_ARG
$  #undef ANSI_ARG
$  #endif
$  #ifdef __STDC__
$  #define ANSI_ARGS(a) a
$  #else
$  #define ANSI_ARGS(a) ()
$  #endif

. -nodebug  - Do not add 
$   #ifndef DEBUG_ALL
$   #define DEBUG_ALL
$   #endif
$ to the wrapper file.
. -anyname   - Generate a single wrapper that can handle the three most common
               cases: trailing underscore, no underscore, and all caps.  The
               choice is based on whether 
$     FORTRANCAPS:       Names are uppercase, no trailing underscore
$     FORTRANUNDERSCORE: Names are lowercase, trailing underscore
$              are defined.  
$     FORTRANDOUBLEUNDERSCORE: Names are lowercase, with TWO trailing
      underscores.  This is needed when some versions of "f2c" are 
      used to generate C for Fortran routines.  Note that f2c uses two
      underscores ONLY when the name already contains an underscore
      (at least on the FreeBSD system that I use that uses f2c).
      To handle this case, the generated code contains the second
      underscore only when the name already contains one.

$     If -mapptr is also chosen, then
$     POINTER_64_BITS
$              will also be used to determine if pointers are to long to
               fit in a 32-bit Fortran integer.  Routines that destroy
               a pointer will need to manually insert a call to 
               __RmPointer.  The routines for managing the pointers are 
	       in ptrcvt.c

	       In addition, if -mpi is used, the MPI profiling names are 
	       also generated, surrounded by MPI_BUILD_PROFILING.

  Note:
  We really need a way to specify a general type as a pointer, so that it
  will be handled as a pointer.  The -mpi option is a kludge for a pressing
  need.  Eventually should provide a "-ptr name" option and keep in a 
  search space when looking for known types.

  Author: 
  Bill Gropp
D*/
main( argc, argv )
int  argc;
char **argv;
{
char routine[MAX_ROUTINE_NAME];
char *infilename;
char outfilename[1024];
char dirname[1024];
char fname[1024], *p;
FILE *fd, *fout, *incfd;
int  nread;
char kind;
char incfile[MAX_FILE_SIZE];
char incbuffer[1024];
int  n_in_file;

/* process all of the files */
strcpy( dirname, "." );
incfile[0]  = 0;
SYArgGetString( &argc, argv, 1, "-dir", dirname, 1024 );
SYArgGetString( &argc, argv, 1, "-I",   incfile, 1024 );
NoFortMsgs		   = SYArgHasName( &argc, argv, 1, "-nomsgs" );
MapPointers		   = SYArgHasName( &argc, argv, 1, "-mapptr" );
if (MapPointers) {
    SYArgGetString( &argc, argv, 1, "-ptrprefix", ptrprefix, 32 );
    }
useFerr			   = SYArgHasName( &argc, argv, 1, "-ferr" );
isMPI			   = SYArgHasName( &argc, argv, 1, "-mpi" );
TranslateVoidStar	   = SYArgHasName( &argc, argv, 1, "-voidisptr" );
MultipleIndirectsAreNative = SYArgHasName( &argc, argv, 1, "-mnative" );
AnsiHeader		   = SYArgHasName( &argc, argv, 1, "-ansiheader" );
AddDebugAll                = SYArgHasName( &argc, argv, 1, "-nodebug" );
IfdefFortranName           = SYArgHasName( &argc, argv, 1, "-anyname" );
if (AnsiHeader) OutputImmed = 0;

if (SYArgHasName( &argc, argv, 1, "-help" )) {
    DoBfortHelp( argv[0] );
    exit( 1 );
    }
/* Open up the file of public includes */    
if (incfile[0]) {
    incfd = fopen( incfile, "r" );
    if (!incfd) {
	fprintf( stderr, "Could not open file %s for -I option\n", incfile );
	}
    }
else
    incfd = 0;

argc--; argv++;
while (argc--) {
    /* Input filename */
    infilename = *argv++;
    fd = fopen( infilename, "r" );
    if (!fd) {
    	fprintf( stderr, "Could not open file %s\n", infilename );
    	continue;
        }
    n_in_file = 0;
    /* Remember file name */
    CurrentFilename = infilename;

    /* Set the output filename */
    SYGetRelativePath( infilename, fname, 1024 );
    /* Strip the trailer */
    p = fname + strlen(fname) - 1;
    while (p > fname && *p != '.') p--;
    *p = 0;
    /* Add an extra h to include files */
    if (p[1] == 'h') {
	p[0] = 'h';
	p[1] = 0;
	}
    if (AnsiHeader) 
	sprintf( outfilename, "%s/%s.ansi", dirname, fname );
    else
	sprintf( outfilename, "%s/%sf.c", dirname, fname );
    /* Don't open the filename yet (wait until we know that we'll have
       some output for it) */
    fout = NULL;

    while (FoundLeader( fd, routine, &kind )) {
	/* We need this test first to avoid creating an empty file, 
	   particularly for initf.c */
	if ((kind == ROUTINE || kind == MACRO) && GetSubClass() == 'C') {
	    if (!NoFortMsgs && !NoFortWarnings) {
		fprintf( stderr, 
			"%s %s(%s) can not be translated into Fortran\n",
			(kind == ROUTINE) ? "Routine" : "Macro", 
			routine, CurrentFilename );
		}
	    SkipText( fd, routine, infilename, kind );
	    continue;
	    }
	if (GetIsX11Routine()) {
	    OutputBuf( &fout, infilename, outfilename, incfd, 
		      "#ifndef TOOLSNOX11\n" );
	    }
	if ((kind == ROUTINE || kind == MACRO) && fout == NULL) {
	    OutputBuf( &fout, infilename, outfilename, incfd, (char *)0 );
	    if (!fout) break;
	    }
	if (IfdefFortranName && fout && routine[0] && 
	    (kind == ROUTINE || kind == MACRO)) {
	    char basename[MAX_ROUTINE_NAME], capsname[MAX_ROUTINE_NAME],
	         nouscorename[MAX_ROUTINE_NAME];
	    OutputUniversalName( fout, routine );
	    }
	if (kind == ROUTINE) {
	    n_in_file++;
	    OutputRoutine( fd, fout, routine, infilename, kind );
	    }
	else if (kind == MACRO) {
	    /* Eventually we can handle this by using the Synopsis to
	       construct an equivalent definition */
	    n_in_file ++;
	    OutputMacro( fd, fout, routine, infilename );
	    }
	else if (kind == INCLUDE) {
	    int guard_x = 0;
	    strcpy( incbuffer, "#include " );
	    /* Grumble.  We'll have to fix this eventually */
	    if (routine[0] != '"' && routine[0] != '<') {
		p = routine + strlen(routine) - 1;
		if (*p == '>')
		    strcat( incbuffer, "<" );
		else if (*p == '"') 
		    strcat( incbuffer, p );
		}
	    /* Special case */
	    /* fprintf( stderr, "include == |%s|\n", routine ); */
	    if (strncmp( routine, "xtools/", 7 ) == 0) {
		guard_x = 1;
		OutputBuf( &fout, infilename, outfilename, incfd, 
			   "#ifndef TOOLSNOX11\n" );
		}
	    strcat( incbuffer, routine );
	    CopyIncludeName( fd, incbuffer + strlen(incbuffer) );
	    strcat( incbuffer, "\n" );
	    OutputBuf( &fout, infilename, outfilename, incfd, incbuffer );
	    if (guard_x) {
		OutputBuf( &fout, infilename, outfilename, incfd, 
			   "#endif\n" );
		if (!fout) break;
		}
	    }
	if (GetIsX11Routine()) {
	    OutputBuf( &fout, infilename, outfilename, incfd, 
		      "#endif\n" );
	    if (!fout) break;
	    }
	}
    fclose( fd );
    if (fout) {
	fclose( fout );
	if (n_in_file == 0) {
	    /* If all we put into the interface file was an include, we delete
	       it */
	    unlink( outfilename );
	    }
	}
    }
return 0;
}

/* We also need to make some edits to the types occasionally.  First, note
   that double indirections are often bugs */
#define MAX_TYPE_NAME 60    /* Maximum length of a type name */
#define MAX_ARGS      128   /* Maximum number of args to a function */
#define MAX_TYPES      64   /* Maximum number of types to a function */

typedef struct {
    char *name;
    int  has_star, is_char, is_native, 
         type,              /* Index into TYPE_LIST array */
         is_FILE, void_function;
    int  implied_star;
    char *fullname;   /* Name with type qualifiers */
    } ARG_LIST;
typedef struct {
    char type[MAX_TYPE_NAME];
    } TYPE_LIST;
typedef struct {
    char name[MAX_TYPE_NAME];
    int  num_stars;
    } RETURN_TYPE;

OutputRoutine( fin, fout, name, filename, kind )
FILE *fin, *fout;
char *name, *filename, kind;
{
int         is_function;
ARG_LIST    args[MAX_ARGS];
TYPE_LIST   types[MAX_TYPES];
RETURN_TYPE rt;
int         nargs, nstrings;
int         ntypes;
int         flag2 = 0;

/* Check to see if this is a C-only routine */
if (GetSubClass() == 'C') {
    if (!NoFortMsgs && !NoFortWarnings) {
	fprintf( stderr, "Routine %s(%s) can not be translated into Fortran\n",
		name, CurrentFilename );
	}
    SkipText( fin, name, filename, kind );
    return;
    }

/* Skip to trailer */
SkipText( fin, name, filename, kind );

/* Get the call to the routine, including finding the argument names */
SkipWhite( fin );
ProcessArgList( fin, fout, filename, &is_function, name, 
	        args, &nargs, &rt, 0 );

SkipWhite( fin );
ProcessArgDefs( fin, fout, args, nargs, types, &ntypes, &nstrings, 0, 
                name, flag2 );
PrintBody( fout, is_function, name, nstrings, nargs, args, types, &rt );
}

/*
    This routine skips the text part of a text page.
 */        
SkipText( fin, name, filename, kind )
FILE *fin;
char *name, *filename;
char kind;
{
int  c;
char lineBuffer[MAX_LINE], *lp;
	
lineBuffer[0] = '+';   /* Sentinal on lineBuffer */
while (1) {
    lp = lineBuffer + 1;
    c  = getc( fin );
    if (c == EOF) break;
    if (c == ARGUMENT || c == VERBATIM)
	SkipLine( fin );
    else if (c == '\n')
	;
    else {
	if (isspace(c) && c != '\n')
	    SkipWhite( fin );
	else 
	    *lp++ = c;
    	/* Copy to end of line; do NOT include the EOL */
    	while ((c = getc( fin )) != EOF && c != '\n') 
    	    *lp++ = c;
    	lp--;
    	while (isspace(*lp)) lp--;
    	lp[1] = '\0';    /* Add the trailing null */
    	if (lineBuffer[1] == kind && strcmp(lineBuffer+2,"*/") == 0)
    	    break;
        }
    }
}

int SkipToSynopsis( fin, kind )
FILE *fin;
char kind;
{
int  c;
char lineBuffer[MAX_LINE], *lp;
	
lineBuffer[0] = '+';   /* Sentinal on lineBuffer */
while (1) {
    lp = lineBuffer + 1;
    c  = getc( fin );
    if (c == EOF) break;
    if (c == ARGUMENT || c == VERBATIM)
	SkipLine( fin );
    else if (c == '\n')
	;
    else {
	if (isspace(c) && c != '\n')
	    SkipWhite( fin );
	else 
	    *lp++ = c;
    	/* Copy to end of line; do NOT include the EOL */
    	while ((c = getc( fin )) != EOF && c != '\n') 
    	    *lp++ = c;
    	lp--;
    	while (isspace(*lp)) lp--;
    	lp[1] = '\0';    /* Add the trailing null */
    	if (lineBuffer[1] == kind && strcmp(lineBuffer+2,"*/") == 0)
    	    break;
	if (lp[0] == ':') {
	    lp = lineBuffer + 1;
	    while (isspace(*lp)) lp++;
	    LowerCase( lp );
	    if (strcmp( lp, "synopsis:" ) == 0) 
		return 1;
	    }
        }
    }
return 0;
}

/* Find the next space delimited token; put the text into token.
   The number of leading spaces is kept in nsp.
   Alpha-numeric tokens are terminated by a non-alphanumeric character
   (_ is allowed in alpha-numeric tokens) */
int FindNextANToken( fd, token, nsp )
FILE *fd;
char *token;
int  *nsp;
{
int fc, c, Nsp;

Nsp = SkipWhite( fd );

fc = c = getc( fd );
if (fc == EOF) return fc;
*token++ = c;
if (isalnum(c) || c == '_') {
    while ((c = getc( fd )) != EOF) {
	if (c != '\n' && (isalnum(c) || c == '_')) *token++ = c;
	else break;
	}
    ungetc( (char)c, fd );
    }
*token++ = '\0';
*nsp     = Nsp;
return fc;
}

OutputBuf( fout, infilename, outfilename, incfd, buffer )
FILE **fout, *incfd;
char *infilename, *outfilename, *buffer;
{
char arch[20];
if (!*fout) {
    *fout = fopen( outfilename, "w" );
    if (!*fout) {
	fprintf( stderr, "Could not open file %s\n", outfilename );
	return;
	}
    fprintf( *fout, "/* %s */\n", infilename );
    if (!AnsiHeader) {
	if (!IfdefFortranName) {
	    SYGetArchType( arch, 20 );
	    fprintf( *fout, "/* Fortran interface file for %s */\n", arch );
	    }
	else {
	    fprintf( *fout, "/* Fortran interface file */\n", arch );
	    }
	/* Turn on the base debugging */
	if (AddDebugAll)
	    fprintf( *fout, "#ifndef DEBUG_ALL\n#define DEBUG_ALL\n#endif\n" );
	if (incfd) {
	    int c;
	    fseek( incfd, 0L, 0 );
	    while ((c = getc( incfd )) != EOF) 
		putc( (char)c, *fout );
	    }
	if (MapPointers) {
	    if (IfdefFortranName) {
		fprintf( *fout, "\n#ifdef POINTER_64_BITS\n\
extern void *%sToPointer();\nextern int %sFromPointer();\n\
extern void %sRmPointer();\n#else\
\n#define %sToPointer(a) (a)\n#define %sFromPointer(a) (int)(a)\
\n#define %sRmPointer(a)\n#endif\n\n", ptrprefix, ptrprefix, ptrprefix, 
			ptrprefix, ptrprefix, ptrprefix );
		}
	    else {
		fprintf( *fout, 
		"extern void *%sToPointer(); extern int %sFromPointer();\n",
			ptrprefix, ptrprefix );
		}
	    }
	}
    }
if (buffer) 
    fputs( buffer, *fout );
}


/* 
   There are a number of things to watch for.  One is that leading blanks are
   considered significant; since the text is being formated, we usually dont
   agree with that. 
 */
OutputMacro( fin, fout, name, filename )
FILE *fin, *fout;
char *name, *filename;
{
int         is_function;
ARG_LIST    args[MAX_ARGS];
TYPE_LIST   types[MAX_TYPES];
RETURN_TYPE rt;
int         nargs, nstrings;
int         ntypes;
int         has_synopsis;
int         c;
int         done;
int         flag2 = 0;

/* Check to see if this is a C-only macro */
if (GetSubClass() == 'C') {
    if (!NoFortMsgs && !NoFortWarnings) {
	fprintf( stderr, "Macro %s(%s) can not be translated into Fortran\n",
		name, CurrentFilename );
	}
    SkipText( fin, name, filename, MACRO );
    return;
    }

/* Skip to the synopsis in the body */
has_synopsis = SkipToSynopsis( fin, MACRO );

done = 0;
if (has_synopsis) {
    /* Get the call to the routine, including finding the argument names */
    SkipWhite( fin );
    /* Process elements of type void * in the following files differently 
       from the default */
    if ((strcmp( filename, "./nonlin/nlfunc_v.h") == 0) ||
       (strcmp( filename, "./nonlin/nlspmat_v.h") == 0) ||
       (strcmp( filename, "./nonlin/nlsles_v.h") == 0)  ||
	TranslateVoidStar)
       flag2 = 1;
    ProcessArgList( fin, fout, filename, &is_function, name, 
		    args, &nargs, &rt, 1 );
    
    SkipWhite( fin );
    done = 
	ProcessArgDefs( fin, fout, args, nargs, types, &ntypes, &nstrings, 1, 
		        name, flag2 );
    PrintBody( fout, is_function, name, nstrings, nargs, args, types, &rt );
    }
else {
    fprintf( stderr, "%s(%s) has no synopsis section\n", 
	     name, CurrentFilename );
    }
/* finish up the section */
if (!done)
    SkipText( fin, name, filename, MACRO );
}

/* Read the arg list and function type */
/* We are moving to being able to suppress generating the output until the
   argument definitions are read */
ProcessArgList( fin, fout, filename, is_function, name, args, Nargs, rt, flag )
FILE        *fin, *fout;
char        *filename, *name;
ARG_LIST    args[512];
RETURN_TYPE *rt;
int         *Nargs;
int         *is_function, flag;
{
int             c;
char            token[1024], *p;
int             i, nsp, bl, leadingm;
static char     rcall[1024];
int             nargs, ln, in_args;
int             reading_function_type = 0;
int             found_name;

SkipWhite( fin );
nargs       = 0;
in_args     = 0;
p           = rcall;
c           = FindNextANToken( fin, p, &nsp );
/* 
   We check for routines that return (functions) versus ones that don't
   by looking for "void".  A special case is functions that return 
   pointers to void; we check for these by looking at the first character
   of the first token after the void.

   We also want to defer generating the function type incase we need to 
   replace a pointer ref with an integer.
 */
strcpy( rt->name, p );
rt->num_stars = 0;
*is_function          = strcmp( p, "void" );
reading_function_type = 1; /* !*is_function; */

if (OutputImmed) {
    for (i=0; i<nsp; i++) putc( ' ', fout );
    }
/* fputs( p, fout ); */
p += strlen( p );
*p++ = ' ';
leadingm = 0;    /* If a newline is encountered before this is one, AND
                    this is a macro, insert one and exit */
found_name = 0;
while (1) {
    c = FindNextANToken( fin, p, &nsp );
    if (c == EOF) {
	fprintf( stderr, "Unexpected EOF in %s\n", filename );
	return;
	}
    if (reading_function_type) {
	if (nsp > 0) strcat( rt->name, " " );
	if (strcmp( p, name ) != 0 && p[0] != '(')
	    strcat( rt->name, p );
	if (c == '*') {
	    *is_function = 1;
	    rt->num_stars++;
	    }
	}
    if (flag && c == '\n' && leadingm == 0) {
	if (OutputImmed)
	    fputs( "()", fout );
	break;
	}
    if (c == '\n') leadingm = 1;
    if (c == '(') {
	reading_function_type = 0;
	if (in_args == 0) {
	    /* Output function type and name */
	    if (rt->num_stars == 0 || !MapPointers) {
		if (useFerr && strncmp( rt->name, "int", 3 ) == 0 ) {
		    if (OutputImmed)
			fputs( "void ", fout );
		    }
		else {
		    if (OutputImmed) {
			fputs( rt->name, fout );
			fputs( " ", fout );
			}
		    }
		}
	    else {
		if (OutputImmed) {
		    fputs( "int ", fout );
		    }
		}
	    if (OutputImmed) 
		OutputRoutineName( name, fout );
	    }
	in_args += 1;
	}
    if (c == ')') {
	in_args -= 1;
	if (in_args == 0) {
	    if (OutputImmed) {
		if (useFerr)
		    fprintf( fout, "%s__ierr ", (nargs > 0) ? ", " : "" );
		fputs( ")", fout ); 
		}
	    break;
	    }
	}
    if (in_args == 0) {
	if (strcmp( p, name ) == 0) {
	    /* Convert to Fortran name.  For now, this just does the
	       lowercase_ version */
	    found_name = 1;
	    }
	else {
	    if (p[0] != '*') 
		fprintf( stderr, "%s:Did not find matching name: %s != %s\n", 
			 filename, p, name );
	    }
	}
    if (in_args == 1) {
	if (c != ',' && c != '(' && c != '\n') {
	    /* Assume that it is a name and remember it */
	    args[nargs].name	     = p;
	    args[nargs].has_star     = 0;
	    args[nargs].implied_star = 0;
	    args[nargs].is_char	     = 0;
	    args[nargs].is_FILE	     = 0;
	    args[nargs].is_native    = 1;   /* Unspecified args are ints */
	    args[nargs].type	     = 0;
	    nargs++;
	    }
	}
    if (OutputImmed) {
	for (i=0; i<nsp; i++) putc( ' ', fout );
	}
    if (in_args) {
	if (OutputImmed)
	    fputs( p, fout );
	p += strlen( p );
	*p++ = 0;
	}
    }

if (!found_name) {
    fprintf( stderr, "%s:Did not find routine name (may be untyped): %s \n", 
 	     filename, name );
    }

/* Handle definitions of the form "type (*Name( args, ... ))()" (this is
   function returns pointer to function returning type). */
SkipWhite( fin );
c = getc( fin );
if (c == '(') {
    SkipWhite( fin );
    c = getc(fin);
    if (c == ')') 
	fputs( "()", fout );
    else 
	ungetc( (char)c, fin );
    }
else 
    ungetc( (char)c, fin );

*Nargs = nargs;

/* If being called from Fortran, we need to append dummy ints for the strings
   passed in.  This requires that we defer to the end of the argument 
   passing the printing of the function declaration line */
/* for (i=0; i<nstrings; i++) fprintf( fout, ",d%d", i ); */

}

/* if flag == 1, stop on empty line rather than { */
/* This needs to distinguish between pointers and values, since all
   parameters are passed by reference in Fortran.  Just to keep things
   lively, there are two ways to indicate a pointer in C:
     type *foo;
     type foo[];

   Needed to add a change that definitions are terminated by ;, not by \n.
 */
int ProcessArgDefs( fin, fout, args, nargs, types, Ntypes, Nstrings, flag, 
		    name, flag2 )
FILE      *fin, *fout;
ARG_LIST  *args;
int       nargs;
TYPE_LIST *types;
int       *Ntypes, *Nstrings, flag, flag2;
char      *name;
{
int      c;
char     token[1024], *p;
int      i, nsp, bl, newline, newstmt;
char     rcall[1024];
int      is_function, in_function;
int      ln, in_args, has_star, implied_star, n_strings, is_char, nstrings,
         is_native, has_array, is_FILE;
int      ntypes, set_void, void_function;
int      done = 0;         /* set to 1 if ate end-of-definition */
int      type_has_star;

newline		  = 1;
newstmt		  = 1;
if (flag) newline = 0;
has_star	  = 0;
type_has_star     = 0;
implied_star	  = 0;
has_array	  = 0;
is_char		  = 0;
is_FILE		  = 0;
nstrings	  = 0;
/* The default type is int */
ntypes		  = 1;
strcpy( types[0].type, "int" );
is_function	  = 0;
in_function	  = 0;
set_void	  = 0;
void_function	  = 0;
while (1) {
    c = FindNextANToken( fin, token, &nsp );
    if (c == EOF || token[0] == '{') break;
    /* Check for empty line; if found, exit.  Otherwise, 
       check for M * / (Macro definition) and handle that case */
    if (flag) {
	if (newline && c == '\n') break;
	if (c == MACRO) {
	    c = getc( fin );
	    if (c == '*') {
		c = getc( fin );
		if (c == '/') {
		    done = 1;
		    break;
		    }
		else { 
		    /* This won't work on all systems. */
		    ungetc( '*', fin );
		    ungetc( (char)c, fin );
		    }
		}
	    else 
		ungetc( (char)c, fin );
	    }
	}

    /* Don't output register */
    if (strcmp( token, "register" ) == 0) continue;
    /* Handle various argument features */
    if (c == '*')                  has_star++;
    else if (c == ',') {
	has_star = type_has_star; has_array = 0; in_function = 0; 
	is_function = 0; set_void = 0; void_function = 0; 
	/* implied_star = 0; */
	}
    else if (c == ';') {
	is_char  = 0; 	is_FILE  = 0; 	has_star = 0; implied_star = 0;
	has_array= 0;   is_native= 0;   is_function = 0; 
	in_function = 0;  set_void = 0; void_function = 0;
	type_has_star = 0;
	newstmt     = 1;
	}
    else if (c == '(') {
         in_function = 1;
         /* If set_void is activated, set the void function indicator */
         if (set_void) {        
            set_void = 0;
            void_function = 1;
            }
         }
    else if (c == ')' && in_function) {
	is_function = 1;
	}
    else if (c == '\n') {
	/* New lines have little meaning in declarations.
	   However, they are necessary to handle blanks lines */
	newline = 1;
	}
#ifdef OLD_CODE
    else if (c == '\n') {
	newline  = 1; 	is_char  = 0; 	is_FILE  = 0; 	has_star = 0;
	has_array= 0;   is_native= 0;   is_function = 0; implied_star = 0;
	in_function = 0;  set_void = 0; void_function = 0;
	}
#endif
    else if (newstmt) {
	if (strcmp( token, "char" ) == 0) is_char = 1;
	if (strcmp( token, "FILE" ) == 0) is_FILE = 1;
	is_native = 0;
	if (strcmp( token, "double" ) == 0 ||
	    strcmp( token, "int"    ) == 0 ||
	    strcmp( token, "float"  ) == 0 ||
	    strcmp( token, "char"   ) == 0 ||
	    strcmp( token, "complex") == 0 ||
	    strcmp( token, "dcomplex")== 0 ||
	    strcmp( token, "MPI_Status") == 0 ||
	    strcmp( token, "BCArrayPart") == 0)
	    is_native = 1;
	/* PETSc types that are implicitly pointers are specified here */
	if (strcmp( token, "XBWindow" ) == 0) {
	    has_star      = 1;
	    type_has_star = 1;
	    implied_star  = 1;
	    }

	if (isMPI) {
	    /* Some things need to be considered ints in the declarations.
	       That is, these are "implicit" pointer objects; often they
	       are pointers to be returned from the calling routine. 
	       These tests set these up as being pointer objects */
	    if (strcmp( token, "MPI_Comm" ) == 0 ||
		strcmp( token, "MPI_Request" ) == 0 ||
		strcmp( token, "MPI_Group" ) == 0 || 
		strcmp( token, "MPI_Intercomm_request" ) == 0 ||    
		strcmp( token, "MPI_Op" ) == 0 ||    
		strcmp( token, "MPI_Uop" ) == 0 ||    
		strcmp( token, "MPI_Datatype" ) == 0) {
 		has_star      = 1;
		type_has_star = 1;
		implied_star  = 1;
		}
	    if (strcmp( token, "MPI_Aint" ) == 0) {
		/* For most systems, MPI_Aint is just long */
 		has_star      = 0;
		type_has_star = 0;
		implied_star  = 0;
		is_native     = 1;
		}
	    }
	if (strcmp( token, "void"   ) == 0) {
           /* Activate set_void only for the files specified by flag2 */
           if (!flag2) is_native = 1;
           else set_void = 1;    
           }
	newline = 0;
	newstmt = 0;
	strcpy( types[ntypes].type, token );
	if (strcmp( token, "struct" ) == 0 || 
	    strcmp( token, "unsigned") == 0) {
	    /* Flush struct to the output */
	    if (OutputImmed) {
		for (i=0; i<nsp; i++) putc( ' ', fout );
		fputs( token, fout );
		}
	    c = FindNextANToken( fin, token, &nsp );
	    strcat( types[ntypes].type, " " );
	    strcat( types[ntypes].type, token );
	    }
	ntypes++;
	}
    else 
	in_function = 0;

    /* Check for "[]".  This won't work for [][3], for example */
    c = getc( fin );
    if (c == '[') {
	has_star++;
	while ((c = getc(fin)) != EOF && c != ']') ;
	has_array = 1;
	}
    else
	ungetc( c, fin );

    /* Look up name */
    for (i=0; i<nargs; i++) {
	if (strcmp( token, args[i].name ) == 0) {
	    args[i].has_star	  = has_star;
	    args[i].implied_star  = implied_star;
	    args[i].is_char	  = is_char;
	    args[i].is_FILE	  = is_FILE;
	    args[i].type	  = ntypes-1;
	    args[i].is_native	  = is_native;
	    args[i].void_function = void_function;
	    if (is_char) nstrings++;
	    if (!has_star) {
		/* This makes it look nicer */
		nsp = 0;
		if (OutputImmed) {
		    for (i=0; i<nsp; i++) putc( ' ', fout );
		    fputs( "*", fout );
		    }
		}
	    else if (has_star > 1) {
		if (!NoFortMsgs && !is_function && 
		     has_star > implied_star + 1) {
		    fprintf( stderr, 
			     "%s(%s) has multiple indirection for %s\n",
			     name, CurrentFilename, args[i].name );
		    }
		if (!MultipleIndirectsAreNative) args[i].is_native = 0;
		}
	    break;
	    }
	}
    if (OutputImmed) {
	for (i=0; i<nsp; i++) putc( ' ', fout );
	fputs( token, fout );
	if (has_array) fputs( "[]", fout );
	}
    }
/* Add the final error return definition */
if (useFerr && OutputImmed) {
    fputs( "int *__ierr;\n", fout );
    }
*Ntypes   = ntypes;
*Nstrings = nstrings;
return done;
}

/*
    Pointer mashing.  There are two kinds of pointer mashing available.
    For systems for which a pointer will fit into an int, we simply
    use the ints to store the pointers.  In this case, a pointer is
    passed to a C routine by using:

    (type *)*(int *)varname

    That is, we convert the varname to an address of an int and deref it.

    On systems with pointers that are longer than ints, we have to do more.
    
    The first step is to convert pointers to indices on input and output
    from the routines.

    The routine __ToPointer converts an index into a pointer.
    The routine __FromPointer converts from a pointer to an index
    The routine __RmPointer   removes a pointer from the table of pointers

    __FromPointer always allocates a new pointer item.
 */
char *ToCPointer( type, name, implied_star )
char *type, *name;
int  implied_star;
{
static char buf[300];

if (MapPointers) 
    sprintf( buf, "\n\t(%s%s)%sToPointer( *(int*)(%s) )", 
	     type, !implied_star ? "* " : "", ptrprefix, name );
else
    sprintf( buf, "\n\t(%s%s)*((int*)%s)", type, !implied_star ? "* " : "", 
	     name );

return buf;
}
/*
   A major question is whether "void *" should be considered the actual
   pointer or an address containing the value of the pointer (the usual "int"
   trick). 

   Since "void *" is used heavily in the communications routines to refer
   to any one of the type double*, int*, ..., I'm going to add void * to
   the list of types that are not translated
 */
PrintBody( fout, is_function, name, nstrings, nargs, args, types, rt )
FILE        *fout;
char        *name;
int         is_function, nstrings, nargs;
ARG_LIST    *args;
TYPE_LIST   *types;
RETURN_TYPE *rt;
{
int  i, j;

/* Known bugs in ansiheader:
   Definitions like     void (*fcn)() fail
   Multiple indirection (char **argv) fail
 */  
if (!OutputImmed) {
    /* Output the function definition */
    if (AnsiHeader) fputs( "extern ", fout );
    fputs( rt->name, fout );
    fputs( " ", fout );
    OutputRoutineName( name, fout );
    if (AnsiHeader) fputs( " ANSI_ARGS(", fout );
    fprintf( fout, "(" );
    for (i=0; i<nargs-1; i++) {
	if (AnsiHeader) {
	    fprintf( fout, "%s", types[args[i].type].type );
	    if (args[i].has_star > 0) {
		for (j=0; j<args[i].has_star; j++) 
		    fputs( "*", fout );
		}
	    fputs( ", ", fout );
	    }
	else 
	    fprintf( fout, "%s, ", args[i].name );
	}
    if (nargs > 0) {
	/* Do the last arg, if any */
	if (AnsiHeader) {
	    fprintf( fout, "%s ", types[args[nargs-1].type].type );
	    if (args[nargs-1].has_star > 0) {
		for (j=0; j<args[nargs-1].has_star; j++) 
		    fputs( "*", fout );
		}
	    }
	else 
	    fprintf( fout, "%s ", args[nargs-1].name );
	}
    else {
	if (AnsiHeader)
	    /* A routine with no arguments gets a 'void' as the argument 
	       name */
	    fputs( "void", fout );
	}
    if (nstrings && !AnsiHeader) {
	for (i=1; i<nstrings; i++) fprintf( fout, ",d%d", i );
	/* Undefined variables are int's by default */
	/* fprintf( fout, "int d0" );
	for (i=1; i<nstrings; i++) fprintf( fout, ",d%d", i );
	fputs( ";\n", fout );
	*/
	}
    fprintf( fout, ")" );
    if (AnsiHeader) {
	/* No more to do */
	fputs( ");\n", fout );
	return;
	}
    else
	fputs( "\n", fout );
    }
fputs( "{\n", fout );
/* Look for special-case translations (currently, "FILE") */
for (i=0; i<nargs; i++) {
    if (args[i].is_FILE) {
	fprintf( fout, "FILE *_fp%d = stdout;\n", i );
	}
    }

/* Generate the routine call with the return */
if (is_function) {
    if (useFerr) {
	fputs( "*__ierr = ", fout );
	}
    else {
	fputs( "return ", fout );
	/* May have to convert type */
	if (MapPointers && rt->num_stars > 0) {
	    /* In this case, we return an integer */
	    fprintf( fout, "%sFromPointer( (void *)( ", ptrprefix );
	    }
	}
    }
fputs( name, fout );
fputs( "(", fout );
for (i=0; i<nargs; i++) {
    if (args[i].is_FILE) 
	fprintf( fout, "_fp%d", i );
    else if (!args[i].is_native && args[i].has_star 
             && !args[i].void_function) {
	if (args[i].has_star == 1 || !MultipleIndirectAreInts) 
	    fprintf( fout, "%s", 
		    ToCPointer( types[args[i].type].type, args[i].name,
			        args[i].implied_star ) );
	else {
	    int j;
	    if (MultipleIndirectsAreNative) {
		fprintf( fout, "%s", args[i].name );
		}
	    else {
		fprintf( fout, "(%s ", types[args[i].type].type );
		if (!args[i].implied_star)
		    for (j = 0; j<args[i].has_star; j++) fputs( "*", fout );
		fprintf( fout, ")*((int *)%s)", args[i].name );
		}
	    }
	}
    else {
	if (!args[i].has_star) 
	    fputs( "*", fout );
	fputs( args[i].name, fout );
	}
    if (i < nargs-1) fputs( ",", fout );
    }
/* fputs( rcall, fout ); */

if (is_function && MapPointers && rt->num_stars > 0 && !useFerr) {
    fprintf( fout, ") )" );
    }
fputs( ");\n}\n", fout );
}

int NameHasUnderscore( p )
char *p;
{
while (*p) 
    if (*p++ == '_') return 1;
return 0;
}

int OutputRoutineName( name, fout )
char *name;
FILE *fout;
{
char buf[256], *p;
int  ln;

p = buf;
strcpy( buf, name );
if (!AnsiHeader) {
    if (IfdefFortranName) {
	LowerCase( p );
	ln = strlen( p );
	p[ln] = '_';
	p[ln+1] = 0;
	}
    else {
#if defined(FORTRANCAPS)
	UpperCase( p );
#elif defined(FORTRANUNDERSCORE)	    
	LowerCase( p );
	ln	= strlen( p );
	p[ln]	= '_';
	p[ln+1]	= 0;
#elif defined(FORTRANDOUBLEUNDERSCORE)
	LowerCase( p );
	ln	= strlen( p );
	if (NameHasUnderscore( p )) {
	    p[ln]	= '_';
	    p[ln+1]	= '_';
	    p[ln+2]	= 0;
	    }
	else {
	    p[ln]	= '_';
	    p[ln+1]	= 0;
	    }
#else
	LowerCase( p );
#endif
	}
    }
fputs( buf, fout );
}

int OutputUniversalName( fout, routine )
FILE *fout;
char *routine;
{
char basename[MAX_ROUTINE_NAME], capsname[MAX_ROUTINE_NAME],
     nouscorename[MAX_ROUTINE_NAME];
strcpy( basename, routine );
LowerCase( basename );
strcat( basename, "_" );
strcpy( capsname, routine );
UpperCase( capsname );
strcpy( nouscorename, routine );
LowerCase( nouscorename );
if (isMPI) {
    if (NameHasUnderscore(nouscorename)) {
	fprintf( fout, "\
#ifdef MPI_BUILD_PROFILING\n\
#ifdef FORTRANCAPS\n\
#define %s P%s\n\
#elif defined(FORTRANDOUBLEUNDERSCORE)\n\
#define %s p%s_\n\
#elif !defined(FORTRANUNDERSCORE)\n\
#define %s p%s\n\
#else\n\
#define %s p%s\n\
#endif\n\
#else\n\
#ifdef FORTRANCAPS\n\
#define %s %s\n\
#elif defined(FORTRANDOUBLEUNDERSCORE)\n\
#define %s %s_\n\
#elif !defined(FORTRANUNDERSCORE)\n\
#define %s %s\n\
#endif\n\
#endif\n\n", basename, capsname, 
	     basename, basename, 
	     basename, nouscorename, 
	     basename, basename, 
	    
	     basename, capsname, 
	     basename, basename, 
	     basename, nouscorename );
	}
    else {
	fprintf( fout, "\
#ifdef MPI_BUILD_PROFILING\n\
#ifdef FORTRANCAPS\n\
#define %s P%s\n\
#elif !defined(FORTRANUNDERSCORE) && !defined(FORTRANDOUBLEUNDERSCORE)\n\
#define %s p%s\n\
#else\n\
#define %s p%s\n\
#endif\n\
#else\n\
#ifdef FORTRANCAPS\n\
#define %s %s\n\
#elif !defined(FORTRANUNDERSCORE) && !defined(FORTRANDOUBLEUNDERSCORE)\n\
#define %s %s\n\
#endif\n\
#endif\n\n", basename, capsname, 
	     basename, nouscorename, 
	     basename, basename, 
	    
	     basename, capsname, 
	     basename, nouscorename );
	}
    }
else {
    if (NameHasUnderscore(nouscorename)) {
	fprintf( fout, "\
#ifdef FORTRANCAPS\n\
#define %s %s\n\
#elif defined(FORTRANDOUBLEUNDERSCORE)\n\
#define %s %s_\n\
#elif !defined(FORTRANUNDERSCORE)\n\
#define %s %s\n\
#endif\n",  basename, capsname, 
	    basename, basename, 
	    basename, nouscorename );
	}
    else {
	fprintf( fout, "\
#ifdef FORTRANCAPS\n\
#define %s %s\n\
#elif !defined(FORTRANUNDERSCORE) && !defined(FORTRANDOUBLEUNDERSCORE)\n\
#define %s %s\n\
#endif\n",  basename, capsname, 
	    basename, nouscorename );
	}
    }
}

int DoBfortHelp( pgm )
char *pgm;
{
fprintf( stderr, "%s - write a Fortran interface to C routines with\n", 
	 pgm );
fprintf( stderr, "routines documented in the `doctext' format\n" );
fprintf( stderr, "Optional arguments:\n" );
fprintf( stderr, "\
filenames - Names the files from which lint definitions are to be extracted\n\
-nomsgs   - Do not generate messages for routines that can not be converted\n\
            to Fortran.\n\
-nofort   - Generate messages for all routines/macros without a Fortran\n\
            counterpart.\n\
-dir name - Directory for output file\n\
-I name   - file that contains common includes\n\
-mapptr   - translate pointers to integer indices\n\
-ferr     - Fortran versions return the value of the routine as the last\n\
            argument (an integer).  This is used in MPI and is a not\n\
            uncommon approach for handling error returns.\n\
-mpi      - Handle MPI datatypes (some things are pointers by definition)\n\
-mnative  - Multiple indirects are native datatypes (no coercion)\n\
-voidisptr - Consider \"void *\" as a pointer to a structure.\n\
-ansiheader - Generate ANSI-C style headers instead of Fortran interfaces\n\
This will be useful for creating ANSI prototypes without ANSI-fying the\n\
code.  The output is in <filename>.ansi .\n\
" );
}
