/* 
   MPI interface

   UNDER CONSTRUCTION

   Pending changes:
   ASYNCRecvId_t should be a pointer to a structure that holds the type and
   source (since these values are returned asynchronously).  A simple heap of
   such structures can be allocated; by using the pointer, they will work 
   even for Fortran users.
 */

#ifndef __commmpi
#define __commmpi

#include "mpi.h"

#ifndef __commh
#include "comm/comm.h"
#endif

extern MPI_Status __STATUS;
extern int __NUMNODES, __MYPROCID, __MPILEN, __FLAG;

/* 
   This won't work for Alpha, but we need to do something... 
   Since MPI uses elements and Chameleon uses bytes to measure length, 
   I've chosen MPI_BYTE as the type for all sends.  This breaks 
   heterogeneous systems, but is a first step 
 */
#ifdef FOO
/* 
   We expected (e.g., in io/pio.c) to use these as constants.  Since
   we are using MPI_BYTE as the type (since lengths are in bytes), 
   the actual values are irrelavent.
 */
#define MSG_INT   (long)MPI_INT
#define MSG_LNG   (long)MPI_LONG
#define MSG_FLT   (long)MPI_FLOAT
#define MSG_DBL   (long)MPI_DOUBLE
#define MSG_OTHER (long)MPI_BYTE
#define MSG_SHRT  (long)MPI_SHORT
#else
#define MSG_INT   0
#define MSG_LNG   1
#define MSG_FLT   2
#define MSG_DBL   3
#define MSG_OTHER 4
#define MSG_SHRT  5
#endif
/*
   The C interface seems to call the Fortran interface so why not use
   the Fortran interface directly. 
*/

/* There are native global ops; I just haven't included them yet */
#define NO_NATIVE_GLOBAL
#define NO_FORCE

	     /*mpc_bsend(buffer,length,to,type);\ */
#define SENDSYNC(type,buffer,length,to,datatype) \
        {LOGBSENDSTART(length,type,to);\
	     MPI_Send( buffer, length, MPI_BYTE, to, type, MPI_COMM_WORLD );\
         LOGBSENDEND(type);}
#define RECVSYNC(type,buffer,length,datatype) {\
        LOGBRECVSTART(length,type);\
        MPI_Recv( buffer, length, MPI_BYTE, MPI_ANY_SOURCE, type, \
		  MPI_COMM_WORLD, &__STATUS );\
        LOGBRECVEND(type,\
     (MPI_Get_count(&__STATUS,MPI_BYTE,&__MPILEN),__MPILEN),__STATUS.MPI_SOURCE);}

#define SENDSYNCNOMEM(type,buffer,length,to,datatype) \
        SENDSYNC(type,buffer,length,to,datatype)
#define RECVSYNCNOMEM(type,buffer,length,datatype)  \
        RECVSYNC(type,buffer,length,datatype);
#define RECVSYNCUNSZ(type,bufp,size,datatype) {\
	MPI_Probe( MPI_ANY_SOURCE, type, MPI_COMM_WORLD, &__STATUS);\
	MPI_Get_count( &__STATUS, MPI_BYTE, &size);\
        bufp=(char*)MALLOC(size);\
        RECVSYNC(type,bufp,size,datatype);}

typedef MPI_Request ASYNCRecvId_t;
typedef MPI_Request ASYNCSendId_t;

#define PIRecvIdNull(a) ((a) == MPI_REQUEST_NULL)
#define PIRecvIdClear(a) a = MPI_REQUEST_NULL
#define PISendIdNull(a) ((a) == MPI_REQUEST_NULL)
#define PISendIdClear(a) a = MPI_REQUEST_NULL

/* Asynchronous routines */
#define SENDASYNC(type,buffer,length,to,datatype,id) \
        {LOGNSENDSTART(length,type,to);\
	     MPI_Isend( buffer,length,MPI_BYTE,to,type,MPI_COMM_WORLD,&(id));\
		 LOGNSENDEND(type);}

/* Design problem:  __STATUS.MPI_SOURCE will be set when the message comes in.
   Thus we really need to modify AsyncRecvId_t to be a structure
   containing the source field! */
#define RECVASYNC(type,buffer,length,datatype,id) \
     { LOGNRECVSTART(length,type);\
     MPI_Irecv( buffer,length,MPI_BYTE,MPI_ANY_SOURCE,type,MPI_COMM_WORLD,&(id));\
     LOGNRECVEND(type);}
#define SENDWAIT(type,buffer,length,to,datatype,id) {\
    LOGWSENDSTART(length,type,to);MPI_Wait(&(id),&__STATUS);\
    LOGWSENDEND(type);}
#define RECVWAIT(type,buffer,length,datatype,id)     \
        {LOGWRECVSTART(type);MPI_Wait(&(id),&__STATUS);\
	     LOGWRECVEND(type,\
	     (MPI_Get_count(&__STATUS,MPI_BYTE,&__MPILEN),__MPILEN),-1);}

#define SENDASYNCNOMEM(type,buffer,length,to,datatype,id) \
        SENDASYNC(type,buffer,length,to,datatype,id)
#define RECVASYNCNOMEM(type,buffer,length,datatype,id) \
        RECVASYNC(type,buffer,length,dataype,id)
#define SENDWAITNOMEM(type,buffer,length,to,datatype,id)  \
        SENDWAIT(type,buffer,length,to,datatype,id)  
#define RECVWAITNOMEM(type,buffer,length,datatype,id)     \
        RECVWAIT(type,buffer,length,datatype,id)
/* does this have Fortran version?? */
#define ASYNCPROBE(type) (MPI_Iprobe(MPI_ANY_SOURCE,type,MPI_COMM_WORLD,\
       &__FLAG, &__STATUS),__FLAG)

#define ASYNCDONE(id)    ( MPI_Test(&(id),&__FLAG,&__STATUS),__FLAG )

#define RECVCANCEL(id)   MPI_Cancel(&(id))
#define SENDCANCEL(id)   MPI_Cancel(&(id))

/* No Forcetype routines.  */
#define SYNCPROBE(type) \
   MPI_Probe( MPI_ANY_SOURCE, type, MPI_COMM_WORLD, &__STATUS )

#define RECVPROBED(type,buffer,length,datatype) \
        RECVSYNC(type,buffer,length,datatype)

#define RECVLEN()  (MPI_Get_count(&__STATUS,MPI_BYTE,&__MPILEN),__MPILEN)
#define RECVFROM() __STATUS.MPI_SOURCE
#define RECVTYPE() __STATUS.MPI_TAG

#define MSGALLOCSEND(msg,maxmsg,type)  \
    msg = (type *)((maxmsg)?MALLOC(maxmsg):0)
#define MSGFREESEND(msg)               if(msg){FREE(msg);}
#define MSGALLOCRECV(msg,maxmsg,type)  \
    msg = (type *)((maxmsg)?MALLOC(maxmsg):0)
#define MSGFREERECV(msg)               if(msg){FREE(msg);}

/* 
   There are global operations, even on subsets!
   This will require some support by procset, since the procset code
   assumed that no-one would provide this functionality.  Procset
   should provide hooks for this, and there should be a procset 
   implementation layer.

   For now, we'll leave support for these operations out.
 */

/* NUMNODES and MYPROCID need to be set by the init routine */
#define NUMNODES __NUMNODES
#define MYPROCID __MYPROCID
#define MSGTYPERANGE(low,high) {void *__out; *low=0; \
		    MPI_Attr_get( MPI_COMM_WORLD, MPI_TAG_UB, &__out ); \
				    *high = *(int*)__out;}
#define MSGSIZES(min,max)      {*min=0;*max=128000;}
#define MSGDISTANCE(from,to)   (from != to)
#define MSGDIAMETER            (1)

#endif


