/*
    These are the definitions particular to the pvm implementation.
 */

#ifndef __commpvm3
#define __commpvm3

#ifndef __commh
#include "comm/comm.h"
#endif

extern int __PVMFROM, __PVMFROMTID, __PVMLEN, __PVMTYPE, __PVMGLOBALTYPE;

#define MSG_INT   0
#define MSG_LNG   1
#define MSG_FLT   2
#define MSG_DBL   3
#define MSG_OTHER 4
#define MSG_SHRT  5

#define NO_ASYNC_SEND
#define NO_ASYNC_RECV
#define NO_FORCE
#define NO_NATIVE_GLOBAL

#ifdef T9000
#ifndef HAVE_PVM_PRECV
#define HAVE_PVM_PRECV
#endif
#ifndef HAVE_PVM_PROBE
#define HAVE_PVM_PROBE
#endif
#include "pvm3.h"
extern int  *PVMtids ;
#define CpvmMapToTid( pid )   PVMtids[pid]
#endif
#if defined(HAVE_PVM_PRECV)
/* 
   There is no obvious way from the pvm header file to determine which 
   version of PVM 3.x this is.  We expect eventually to use configure
   to determine this.  Note also that the "datatype" argument to 
   pvm_psend and pvm_precv are NOT USED for conversions, just for length
   (at least in 3.3.0).
*/
#define SENDSYNC(type,buffer,length,to,datatype) \
        {LOGBSENDSTART(length,type,to);\
	pvm_psend(CpvmMapToTid(to),type,(char*)(buffer),length,PVM_BYTE);\
        LOGBSENDEND(type);}
#define RECVSYNC(type,buffer,length,datatype) { \
        __PVMFROM = -1;__PVMTYPE = type; __PVMLEN = length;\
	LOGBRECVSTART(length,type);\
        pvm_precv(-1,type,(char*)(buffer),length,PVM_BYTE,\
		  &__PVMFROMTID,&__PVMLEN,&__PVMLEN); \
        LOGBRECVEND(type,__PVMLEN,__PVMFROM);}
/* Note: if the buffer was smaller than the received message, we should
   generate an error */
#define RECVSYNCUNSZ(type,bufp,size,datatype) {\
      	__PVMFROM = -1;__PVMTYPE = type; __PVMLEN = size;\
	LOGBRECVSTART(-1,type);\
      Cpvm_recv_unsz(&__PVMTYPE,&__PVMFROM,(int**)&(bufp),&__PVMLEN,datatype);\
        LOGBRECVEND(type,size,__PVMFROM);}

#else
#define SENDSYNC(type,buffer,length,to,datatype) \
        {LOGBSENDSTART(length,type,to);\
	Cpvm_send(type,to,(int*)(buffer),length,datatype);LOGBSENDEND(type);}
#define RECVSYNC(type,buffer,length,datatype) { \
        __PVMFROM = -1;__PVMTYPE = type; __PVMLEN = length;\
	LOGBRECVSTART(length,type);\
        Cpvm_recv(&__PVMTYPE,&__PVMFROM,(int*)(buffer),&__PVMLEN,datatype); \
        LOGBRECVEND(type,__PVMLEN,__PVMFROM);}
/* Note: if the buffer was smaller than the received message, we should
   generate an error */
#define RECVSYNCUNSZ(type,bufp,size,datatype) {\
      	__PVMFROM = -1;__PVMTYPE = type; __PVMLEN = size;\
	LOGBRECVSTART(-1,type);\
      Cpvm_recv_unsz(&__PVMTYPE,&__PVMFROM,(int**)&(bufp),&__PVMLEN,datatype);\
        LOGBRECVEND(type,size,__PVMFROM);}

#endif

#if defined(HAVE_PVM_PROBE)
#define SYNCPROBE(type) {while(pvm_probe(-1,type)==0);}
#define ASYNCPROBE(type) (pvm_probe(-1,type) > 0)
#else
#define SYNCPROBE(type) {while(Cpvm_probe(type)==-1);}
#define ASYNCPROBE(type) (Cpvm_probe(type) != -1)
#endif

#define SENDSYNCNOMEM(type,buffer,length,to,datatype) \
        SENDSYNC(type,buffer,length,to,datatype)
#define RECVSYNCNOMEM(type,buffer,length,datatype)  \
        RECVSYNC(type,buffer,length,datatype)

/* I'm not sure that this is correct */
#define RECVPROBED(type,buffer,length,datatype) \
        RECVSYNC(type,buffer,length,datatype)

#define RECVLEN()  __PVMLEN
#define RECVFROM() CpvmMapFromTid(__PVMFROMTID)
#define RECVTYPE() __PVMTYPE

#define MSGALLOCSEND(msg,maxmsg,type)  \
    msg = (type *)((maxmsg)?MALLOC(maxmsg):0)
#define MSGFREESEND(msg)               if(msg){FREE(msg);}
#define MSGALLOCRECV(msg,maxmsg,type)  \
    msg = (type *)((maxmsg)?MALLOC(maxmsg):0)
#define MSGFREERECV(msg)               if(msg){FREE(msg);}

#define NUMNODES __NUMNODES
#define MYPROCID __MYPROCID

/* The max value below is approximate */
#define MSGTYPERANGE(low,high) {*(low)=0;*(high)=10000000;}
#define MSGDISTANCE(from,to)   gdistdf(from,to)
#define MSGSIZES(min,max)      {*min=0;*max=(16384*8096);}
#define MSGDIAMETER            (NUMNODES-1)

#endif

