C        A simple program to solve u_xx = 12x^2 on a domain
C        from 0 to 1. With u(x=0) = 0; u(x=1) = 1;

       subroutine Mult(N,x,y)
       integer    N, i
       double precision x(*), y(*)
#include "comm/fcomm.h"
       do 10 i=2,N-1
         y(i) = 2.0*x(i) - x(i-1) - x(i+1)
 10    continue
       if (PImytid() .eq. 0) then 
         y(1) = 2.0*x(1) - x(2)
       else
         call PIbsend(1,y(2),dblsze,PImytid()-1,MSG_DBL)
         call PIbrecv(2,y(1),dblsze,MSG_DBL)
       endif 
       if (PImytid() .eq. PInumtids()-1) then 
         y(N) = 2.0*x(N) - x(N-1)
       else 
         call PIbsend(2,y(N-1),dblsze,PImytid()+1,MSG_DBL)
         call PIbrecv(1,y(N),dblsze,MSG_DBL)
       endif 
       return 
       end 

       subroutine Pre(ptype,N,x,y)
       integer  N,ptype
       double  precision x(*),y(*)
       if (ptype .eq. 0) then
         call JacobiPre(N,x,y)
       else if (ptype .eq. 1) then
         call GSPre(N,x,y)
       endif
       return
       end

       subroutine GSPre(N,x,y)
       integer    N,i,i1,i2
       double precision x(*), y(*), z(10000)
#include "comm/fcomm.h"
     
       i1 = 1
       if (PImytid() .gt. 0) i1 = 2 
       i2 = N 
       if (PImytid() .lt. PInumtids()-1) i2 = N - 1 
       z(i1) = .5*x(i1)
       do 10 i=i1+1,i2
         z(i) = .5*(x(i) + z(i-1))
 10    continue
       y(i2) = .5*z(i2)
       do 30 i=i2-1,1,-1
         y(i) = .5*(z(i) + y(i+1))
 30    continue
       if (PImytid() .gt. 0) then 
         call PIbsend(1,y(2),dblsze,PImytid()-1,MSG_DBL)
         call PIbrecv(2,y(1),dblsze,MSG_DBL)
       endif
       if (PImytid() .lt. PInumtids()-1) then 
         call PIbsend(2,y(N-1),dblsze,PImytid()+1,MSG_DBL)
         call PIbrecv(1,y(N),dblsze,MSG_DBL)
       endif 
       return 
       end

       subroutine JacobiPre(N,x,y)
       integer    N,i
       double precision x(*), y(*)
       do 10 i=1,N
         y(i) = x(i)
 10    continue
       return 
       end

       subroutine worker
#include "xtools/fbasex11.h"
#include "comm/fcomm.h"
#include "system/fsystem.h"
       integer  usex,N, i, its, win1, win2, win3, CGSolve,ptype
       integer  offset
       double precision  f(10000),u(10000),ustar(10000),tol,h,dp
       double precision  x(10000),pi,xstart,time

       if (PImytid() .eq. 0) then
         print*,'Enter N'
         read*,N
         print*,'Enter 1 for XWindows, 0 for not'
         read*,usex 
         print*,'Enter 1 for Gauss-Seidel 0 for Jacobi'
         read*,ptype 
       endif 
       Call PIbcastSrc(N , intsze, 0, PSAllProcs, MSG_INT)
       Call PIbcastSrc(usex , intsze, 0, PSAllProcs, MSG_INT)
       Call PIbcastSrc(ptype , intsze, 0, PSAllProcs, MSG_INT)

       tol = 1.e-6
       pi = 3.14159265358979

       if (usex .eq. 1) then
         win1 = XBOpenPerProcessorWindow('Solution',0)
         win2 = XBOpenPerProcessorWindow('Residual',1)
         win3 = XBOpenPerProcessorWindow('Error',2)
       else
         win1 = 0
       endif 
       h = 1./(N+1) 
       nn     = (N + PInumtids() - 1)/PInumtids()
       offset = nn*PImytid()
       if ( (nn + offset) .gt. N ) then
         nn = N - offset
       endif
       if (PImytid() .gt. 0) nn = nn + 1
       if (PImytid() .lt. PInumtids()-1) nn = nn + 1
       if (PImytid() .gt. 0) then
         xstart = h*offset
       else
         xstart = h
       endif

       if (nn .gt. 10000) then
         print*, 'Recompile lapf.F and cgf.F with larger allocations'
         call SYexitall(0,0) 
       endif

       x(1) = xstart
       do 10 i=2,nn
         x(i) = x(i-1) + h
 10    continue
       do 20 i=1,nn
         f(i)     = -12.d0*h*h*x(i)*x(i) 
         u(i)     = 0.0
         ustar(i) = x(i)**4.0
 20    continue
       if (PImytid() .eq. PInumtids()-1) f(nn) = f(nn) + 1.0

       call PIgsync(PSAllProcs)
       time = SYGetElapsedTime()
       its = CGSolve(nn,f,u,N,tol,win1,win2,win3,x,ustar,ptype)
       time = SYGetElapsedTime() - time
       call DVaxpy( nn, -1.d0, u,ustar )
       call DVnorm( nn, ustar, dp)
       if (PImytid() .eq. 0) then
         print*,'Time ',time,' Its ',its, ' Error ',dsqrt(h)*dp
       endif
       return 
       end

       subroutine GraphSolution(win,x,u,n)
       integer win, n, i, lg
       double precision x(*), u(*), umin, umax, work
#include "xtools/fbasex11.h"
#include "comm/fcomm.h"

       lg = XBLineGraphCreate(win, 1)
       do 10 i=1,n
         call XBLineGraphAddPoint(lg,x(i),u(i))
 10    continue
       umin = 1.d8
       umax = -1.d8
       do 20 i=1,n
         if (umin .gt. u(i)) umin = u(i)
         if (umax .lt. u(i)) umax = u(i)
 20    continue
       call PIgdmax(umax,1,work,PSAllProcs)
       call PIgdmin(umin,1,work,PSAllProcs)
       call XBLineGraphSetLimits(lg,x(1),x(n),umin,umax)
       call XBLineGraphDraw(lg)
       call XBLineGraphDestroy(lg)
       return 
       end
C        The Conjugate gradient code 
C
      integer function  CGSolve(N,B,U,mit,tol,win1,win2,win3,x,ustar,pt)
#include "xtools/fbasex11.h"
      integer N,mit,win1,win2,win3,pt,i
      double  precision B(N),U(N),tol,x(N),ustar(N)
      double precision  a,beta,betaold,bb,dp,Z(10000),R(10000)
      double precision  P(10000), W(10000)
  
      call Mult(N,U,R)
      call DVaypx(N, -1.0, B, R)
      call Pre(pt,N,R,Z)

      do 10 i=0,mit-1
        CGSolve = i
        call DVdot( N, R, Z, beta)
        if (i .eq. 0) then
          if (beta .eq. 0.0) return 
          call DVcopy( N, Z, P )
        else 
          bb = beta/betaold
          call DVaypx( N, bb, Z, P )
        endif 
        betaold = beta
        call Mult(N,P,Z)
        call DVdot( N, P, Z, dp)
        a = beta/dp
        call DVaxpy( N, a, P,U )
        call DVaxpy( N, -a, Z,R )
        call Pre(pt,N,R,Z)
        call DVnorm( N, Z, dp)
        if (win1 .ne. 0) then 
          call GraphSolution(win1,x,U,N)
          call GraphSolution(win2,x,R,N)
          call DVcopy(N,U,W)
          call DVaxpy(N, -1.d0, ustar,W )
          call GraphSolution(win3,x,W,N)
          print*,'It ',i,' Residual ',dp 
        endif 
        if (dp .lt. tol) return 
 10   continue  
      return 
      end
C
C             The vector operations 
C
       subroutine DVdot( N, x, y, z )
       integer N,i,i1,i2
       double precision x(*), y(*), z,work
#include "comm/fcomm.h"
       z = 0.0
       if (PImytid() .gt. 0) then
         i1 = 2
       else
         i1 = 1
       endif 
       if (PImytid() .lt. PInumtids()-1) then
         i2 = N-1
       else
         i2 = N
       endif
       do 10 i=i1,i2
         z = z + x(i)*y(i)
 10    continue
       call PIgdsum(z,1,work,PSAllProcs)
       return 
       end
   
       subroutine DVnorm( N, x, z )
       integer N,i,i1,i2
       double precision x(*), z,work
#include "comm/fcomm.h"
       z = 0.0
       if (PImytid() .gt. 0) then
         i1 = 2
       else
         i1 = 1
       endif
       if (PImytid() .lt. PInumtids()-1) then
         i2 = N-1
       else
         i2 = N
       endif
       do 10 i=i1,i2
         z = z + x(i)*x(i)
 10    continue
       call PIgdsum(z,1,work,PSAllProcs)
       z = dsqrt(z)
       return 
       end
   
       subroutine DVcopy( N, x, y )
       integer N,i
       double precision x(*), y(*)
       do 10 i=1,N
         y(i) = x(i)
 10    continue
       return 
       end
       
       subroutine DVaxpy( N, alpha, x, y )
       integer N,i
       double precision alpha, x(*), y(*)
       do 10 i=1,N
         y(i) = alpha*x(i) + y(i)
 10    continue
       return 
       end
       
       subroutine DVaypx( N, alpha, x, y )
       integer N,i
       double precision alpha, x(*), y(*)
       do 10 i=1,N
         y(i) = x(i) + alpha*y(i)
 10    continue
       return 
       end
