#ifndef lint
static char vcid[] = "$Id: gcolx.c,v 1.3 1994/06/16 20:22:47 gropp Exp $";
#endif

#include "comm/comm.h"
#include "comm/procset.h"
#include "comm/global/global.h"
#include <stdio.h>

/* #define DEBUG */
extern void PIcor();

/* 
   This is gcol but with sizes known in advance.  The algorithm is to
   find the location in gbuf for the local contribution and then 
   merge the data together.  This is done by ORing the bits.
 */
void gcolxsetBM( lbuf, gsizes, gbuf, procset, datatype )
char    *lbuf, *gbuf;
int     *gsizes, datatype;
ProcSet *procset;
{
int  size, i, lidx, np, loc;
char *work;


/* Find my position */
lidx = PSMYPROCID(procset);
np   = PSNUMNODES(procset);
loc  = 0;
size = 0;
for (i=0; i<lidx; i++) 
    loc += gsizes[i];
size = loc;
while (i < np)
    size += gsizes[i++];

MEMSET(gbuf,0,size);
MEMCPY(gbuf+loc,lbuf,gsizes[lidx]);
work = (char *) MALLOC( size );    CHKPTR(work);
/* Just do OR reduction of vector gbuf.  Later we should consider
   something more efficient (at least for large collections).  */
(*PIcomb)( gbuf, size, work, procset, sizeof(char), datatype, PIcor );
FREE( work );
}

/* Here is a different version that uses an algorithm that sends less data 
   and does not have a single choke point.  The algorithms is
   exchange my current buffer with my partner, who is either 
   immediately above or below me in the.  This allows me to send
   contiguous data.  One potential problem: the choice of partner may
   generate collisions in the network.  We'll let a different algorithm
   deal with that.

   This algorithm is complicated by having to deal with non-powers-of-two
   nodes.  These are handled by using recursive subdivision:
   Find the largest contained power of two (N) in the full set (M) and the
   largest power of two in the remainder (n) (recursively)

   Perform the steps of the algorithm until the exchange distance is n.
   At this point, pause, and have processors N-n to N-1 receive from 
   processors N to N+n-1 (zero-origin).  Complete the algorithm.
   As the last step, have processors N-(M-N) to N-1 send to processors
   N to M-1 (the last exchange step).  Note that this last step is NOT
   handled recursively; only the bottom set of nodes performs the 
   communication.  This is important in maintaining the performance of the
   code to ceiling(log(p)).

   Because of the recursive nature of this algorithm, I've split it into
   a few pieces.  These are

 */

/* Take n and return Nsmall <= n <= Nbig, with Nsmall and Nbig powers-of-two */
void PIiFindTwo( n, Nbig, Nsmall )
int n, *Nbig, *Nsmall;
{
int mask, lmask;
mask = 0x1;
lmask = 0;
while (mask < n) {
    lmask = mask;
    mask  <<= 1;
    }
*Nbig   = mask;
if (mask == n) lmask = mask;
*Nsmall = lmask;
}

void gcolxsetEX( lbuf, gsizes, gbuf, procset, datatype )
char    *lbuf, *gbuf;
int     *gsizes, datatype;
ProcSet *procset;
{
int  size, i, lidx, np, loc;
char *work;
int  dist, nbr, nloc, cursize, nbrtid, etag, lphase, cnt, mask, inbr, nsize;

/* Find my position */
lidx = PSMYPROCID(procset);
np   = PSNUMNODES(procset);
loc  = 0;
for (i=0; i<lidx; i++) 
    loc += gsizes[i];
size = loc;
while (i < np)
    size += gsizes[i++];
cursize = gsizes[lidx];

if (!procset) {
    _PIPHASE= _PIPHASE ? 0 : 1;
    lphase  = _PIPHASE << MSG_PHASE_SHFT;
    }
else {
    GMSGPHASE(procset,lphase);
    }
etag = GMSGTYPE(procset,MSG_UP|lphase);

MEMCPY(gbuf+loc,lbuf,gsizes[lidx]);
dist = 0x1;
cnt  = 0;
mask = ~0;
while (dist < np) {
    nbr = lidx ^ dist;
    /* Exchange data with neighbor.  To avoid problems with head-to-head,
       we order the exchanges */
    if (nbr >= 0 && nbr < np) {
	nbrtid = PStidFromRank( procset, nbr );
	if (lidx & dist) {
	    /* nbr is below in the tree */
	    PIbsend( etag, gbuf+loc, cursize, nbrtid, datatype );
	    nloc = 0;
	    /* inbr gives us the neighbor's current index; that is, the
	       lowest ranked processor that the NEIGHBOR has already exchanged
	       with */
	    inbr = nbr & mask;
	    for (i=0; i<inbr; i++) 
		nloc += gsizes[i];
	    PIbrecv( etag, gbuf+nloc, size-nloc, datatype );
#ifdef DEBUG
/*
	    printf( "[%d]-%d(high) exchanging %d with %d\n", 
		    PImytid, dist, cursize, nbrtid );
 */
	    if (PIsize() != loc - nloc) {
		fprintf( stderr, "[%d] got wrong size (%d) from %d(%d)\n", 
			 lidx, PIsize(), nbrtid, PIfrom() );
		}
	    if (PIfrom() != nbrtid) {
		fprintf( stderr, "[%d] got message from wrong nbr %d(%d)\n", 
			 lidx, nbrtid, PIfrom() );
		}
#endif
	    cursize += loc - nloc;
	    loc	     = nloc;
	    }
	else {
	    /* nbr is immediately above us */
	    nloc = loc + cursize;
	    PIbrecv( etag, gbuf+nloc, size-nloc, datatype );
	    nsize = PIsize();
#ifdef DEBUG
/*
	    printf( "[%d]-%d(low)  exchanging %d with %d\n", 
		    PImytid, dist, cursize, nbrtid );
 */
	    if (PIfrom() != nbrtid) {
		fprintf( stderr, "[%d] got message from wrong nbr %d(%d)\n", 
			 lidx, nbrtid, PIfrom() );
		}
#endif
	    PIbsend( etag, gbuf+loc, cursize, nbrtid, datatype );
	    cursize += nsize;
	    }
	}
    dist <<= 1;
    mask <<= 1;
    etag --;
    }
}


/* Exchange code for given dimensions
   Input Parameters:
.  Dist - Pointer to current distance of exchanges
.  Mask - Pointer to current mask (see algorithm)
.  Etag - Pointer to current message tag value
.  final - 2*size of final exchange ( = number of processes this will
           complete the exchanges for)
.  lidx  - My local index (rank) in the set
.  gbuf  - global buffer
.  loc - gbuf[loc] is bottom of filled-in part
.  cursize - gbuf[loc] to gbuf[loc+cursize-1] are filled in
.  datatype - Type of data
.  lmap - lmap[i] is the absolute task id of the i'th process.  If lmap
   is null, use i == lmap[i].
 */
void PIiSubsetEX( Dist, Mask, final, Loc, Cursize, gsizes, 
		  lidx, Etag, gbuf, size, datatype, 
		  lmap )
int     *Dist, *Mask, *Loc, *Cursize, final, lidx, *Etag, size, datatype, 
        *lmap, *gsizes;
char    *gbuf;
{
int nbr, nbrtid, nloc, inbr, etag = *Etag, i, nsize;
int mask = *Mask, dist = *Dist, cursize = *Cursize, loc = *Loc;
int np = final; /* ??? */

while (dist < final) {
    nbr = lidx ^ dist;
    /* Exchange data with neighbor.  To avoid problems with head-to-head,
       we order the exchanges.  This needs to be in the
       subcube? */
    if (nbr >= 0 && nbr < np) {
	nbrtid = (lmap ? lmap[nbr] : nbr);
	if (lidx & dist) {
	    /* nbr is below in the tree */
	    PIbsend( etag, gbuf+loc, cursize, nbrtid, datatype );
	    nloc = 0;
	    /* inbr gives us the neighbor's current index; that is, the
	       lowest ranked processor that the NEIGHBOR has already exchanged
	       with */
	    inbr = nbr & mask;
	    for (i=0; i<inbr; i++) 
		nloc += gsizes[i];
	    PIbrecv( etag, gbuf+nloc, size-nloc, datatype );
#ifdef DEBUG
/*
	    printf( "[%d]-%d(high) exchanging %d with %d\n", 
		    PImytid, dist, cursize, nbrtid );
 */
	    if (PIsize() != loc - nloc) {
		fprintf( stderr, "[%d] got wrong size (%d) from %d(%d)\n", 
			 lidx, PIsize(), nbrtid, PIfrom() );
		}
	    if (PIfrom() != nbrtid) {
		fprintf( stderr, "[%d] got message from wrong nbr %d(%d)\n", 
			 lidx, nbrtid, PIfrom() );
		}
#endif
	    cursize += loc - nloc;
	    loc	     = nloc;
	    }
	else {
	    /* nbr is immediately above us */
	    nloc = loc + cursize;
	    PIbrecv( etag, gbuf+nloc, size-nloc, datatype );
	    nsize = PIsize();
#ifdef DEBUG
/*
	    printf( "[%d]-%d(low)  exchanging %d with %d\n", 
		    PImytid, dist, cursize, nbrtid );
 */
	    if (PIfrom() != nbrtid) {
		fprintf( stderr, "[%d] got message from wrong nbr %d(%d)\n", 
			 lidx, nbrtid, PIfrom() );
		}
#endif
	    PIbsend( etag, gbuf+loc, cursize, nbrtid, datatype );
	    cursize += nsize;
	    }
	}
    dist <<= 1;
    mask <<= 1;
    etag --;
    }
*Etag    = etag;
*Cursize = cursize;
*Loc     = loc;
*Mask    = mask;
*Dist    = dist;
}

#ifdef FOO

void gcolxsetEX2( lbuf, gsizes, gbuf, procset, datatype )
char    *lbuf, *gbuf;
int     *gsizes, datatype;
ProcSet *procset;
{
int  size, i, lidx, np, loc;
char *work;
int  dist, nbr, nloc, cursize, nbrtid, etag, lphase, cnt, mask, inbr, nsize;
int  NPbig, NPsmall,         /* Size of local piece */
     NPsmaller,              /* Size of largest power-of-two piece in 
				remainder */
     NLeftover;

/* Find my position */
lidx = PSMYPROCID(procset);
np   = PSNUMNODES(procset);
loc  = 0;
for (i=0; i<lidx; i++) 
    loc += gsizes[i];
size = loc;
while (i < np)
    size += gsizes[i++];
cursize = gsizes[lidx];

if (!procset) {
    _PIPHASE= _PIPHASE ? 0 : 1;
    lphase  = _PIPHASE << MSG_PHASE_SHFT;
    }
else {
    GMSGPHASE(procset,lphase);
    }
etag = GMSGTYPE(procset,MSG_UP|lphase);

MEMCPY(gbuf+loc,lbuf,gsizes[lidx]);

PIiFindTwo( np, &NPbig, &NPsmall );
NLeftover = np - NPsmall;
if (lidx >= NPsmall) {
    /* Call recursively */
    /* ... */
    /* Send data back to parent */
    if (lidx - NPsmall < NPloc)
	PIbsend( ... );
    /* Everyone waits for value back */
    PIbrecv( ... );
    return;
    }

/* Do the exchanges until d == NPsmall */
dist = 1;
mask = ~0;
PIiSubsetEX( &dist, &mask, NPsmaller, &loc, &cursize, gsizes, lidx, &etag
	     gbuf, size, datatype, procset ? procset->node_nums : (int *)0 );
/* Get data from the upper piece */
/* ... */
/* Complete the exchanges */
PIiSubsetEX( &dist, &mask, NPsmall, &loc, &cursize, gsizes, lidx, &etag
	     gbuf, size, datatype, procset ? procset->node_nums : (int *)0 );
/* Send data to the leftovers */
if (lidx > NPsmall - Nleftover) {
    PIbsend( ... );
    }
}
#endif

static void (*GCOLXFCN)() = gcolxsetBM;

void gcolxset( lbuf, gsizes, gbuf, procset, datatype )
char    *lbuf, *gbuf;
int     *gsizes, datatype;
ProcSet *procset;
{
(*GCOLXFCN)( lbuf, gsizes, gbuf, procset, datatype );
}

void PISetGcolxFunc( func )
void (*func)();
{
GCOLXFCN = func;
}
