#include "tools.h"
#include "system/system.h"
#include "comm/comm.h"
#include "comm/io/pio.h"
#include <math.h>

main( argc, argv )
int  argc;
char **argv;
{
int worker();
PICall( worker, argc, argv );
}

int worker( argc, argv )
int  argc;
char **argv;
{
int         nx, ny, nr, nc, myid, ri, rj, i, j, g;
int         nxin, nyin, gin;
int         px, py;    /* Size of partition in each dimension */
PIFILE      *fp;
PIArrayPart sz[2];
int         nd;
int         errcnt;
double      *v, *v2;
double      t1, t2;
int         bytes;
int         has_unitree;
char        fname[1024];

/* First, make unitree available as a transport */
has_unitree = PIOAddUnitree();

/* get the global size */
if (SYArgHasName( &argc, argv, 1, "-help" )) {
    fprintf( stderr, "%s -n nx -m ny -g n\n", argv[0] );
    return 0;
    }
nx = 16;
ny = 16;
g  = 0;
SYArgGetInt( &argc, argv, 1, "-n", &nx );
SYArgGetInt( &argc, argv, 1, "-m", &ny );
SYArgGetInt( &argc, argv, 1, "-g", &g );
if (SYArgHasName( &argc, argv, 1, "-block" ))
     PIFSetWritersBlock();

strcpy( fname, "foo" );
SYArgGetString( &argc, argv, 1, "-name", fname, 1024 );

if (SYArgHasName( &argc, argv, 1, "-help" )) {
    fprintf( stderr, "\
example1 [-n nx] [-m ny] [-g n] [-name filename] [-block]\n\
Write an array of global size nx x ny, with ghost size n, to file given\n\
with -name.  If -block used, then a different (PIFSetWritersBlock)\n\
algorithm is used to collect data.  You may use unitree names if unitree\n\
is available\n\
The created file is written in PIO_AS_SEQUENTIAL mode\n" );
    return 1;
    }

fp = PIFopen( fname, PIAllProcs, O_WRONLY | O_CREAT, PIO_AS_SEQUENTIAL, -1 );
PIWriteCommon( fp, "%4d",   4, &nx, 1, MSG_INT );
PIWriteCommon( fp, "%4d",   4, &ny, 1, MSG_INT );
PIWriteCommon( fp, "%4d\n", 5, &g,  1, MSG_INT );

nd   = 2;
nr   = PInumtids;
myid = PImytid;
if (nr > 1) 
    nr = sqrt( (double)nr + 0.1 );
nc = PInumtids / nr;
if (nr * nc != PInumtids) {
    fprintf( stderr, "Test only works for square arrays of processors\n" );
    return;
    }
ri = myid % nc;
rj = myid / nc;
    
px = nx / nc;
py = ny / nr;
/* printf( "[%d](%d,%d)\n", PImytid, ri, rj ); */
/* 
   The GLOBAL array is A(0:nx-1,0:ny-1).  The local array is
   declared as B(0:,0:), with B(g:,g:) matching A(ri*(nx/nc):,rj*(ny/nr):), 
   where the position of the processor in the mesh is given by (ri,rj)
 */
sz[0].mdim   = nx;
sz[0].ndim   = px + 2 * g;
sz[0].start  = g;
sz[0].end    = px - 1 + g;
sz[0].gstart = ri * px;
sz[0].gend   = sz[0].gstart + sz[0].ndim - 1;

sz[1].mdim   = ny;
sz[1].ndim   = py + 2 * g;
sz[1].start  = g;
sz[1].end    = py - 1 + g;
sz[1].gstart = rj * py;
sz[1].gend   = sz[1].gstart + sz[1].ndim - 1;
/* printf( "[%d]<%d,%d>\n", PImytid, sz[0].gstart, sz[1].gstart );  */
fflush(stdout);
v = (double *)MALLOC( sz[0].ndim * sz[1].ndim * sizeof(double) );  CHKPTR(v);
for (j=sz[1].start; j<=sz[1].end; j++) {
    for (i=sz[0].start; i<=sz[0].end; i++) {
	v[i + sz[0].ndim * j] = 
	    i + sz[0].gstart + (j + sz[1].gstart) * sz[0].mdim;
	}
    }

PIgsync( PIAllProcs );
t1 = SYGetElapsedTime();
PIWriteDistributedArray( fp, "%12.4e\n", 13, sz, nd, v, MSG_DBL );
PIgsync( PIAllProcs );
t2 = SYGetElapsedTime() - t1;
if (PImytid == 0) {
    bytes = 13 * nx * ny;
    printf( "Wrote %d bytes at %f Mbytes/second\n", bytes, bytes/(1.0e6*t2) );
    fflush( stdout );
    }
PIFclose( fp );

/* Now, try to READ the data back; check that it is correct */
fp = PIFopen( fname, PIAllProcs, O_RDONLY, PIO_AS_SEQUENTIAL, -1 );
if (!fp) {
    fprintf( stderr, "Can not open file for input\n" );
    return 0;
    }
v2 = (double *)MALLOC( sz[0].ndim * sz[1].ndim * sizeof(double) );  CHKPTR(v2);

PIReadCommon( fp, "%4d",   4, &nxin, 1, MSG_INT );
PIReadCommon( fp, "%4d",   4, &nyin, 1, MSG_INT );
PIReadCommon( fp, "%4d\n", 5, &gin,  1, MSG_INT );

PIgsync( PIAllProcs );
t1 = SYGetElapsedTime();
PIReadDistributedArray( fp, "%12le\n", 13, sz, nd, v2, MSG_DBL );
PIgsync( PIAllProcs );
t2 = SYGetElapsedTime() - t1;
if (PImytid == 0) {
    bytes = 13 * nx * ny;
    printf( "Read %d bytes at %f Mbytes/second\n", bytes, bytes/(1.0e6*t2) );
    fflush( stdout );
    }
PIFclose( fp );

/* Check that the values are correct */
if (nxin != nx) 
    fprintf( stderr, "NX is incorrect in file (got %d expected %d)\n", 
	     nxin, nx );
if (nyin != ny) 
    fprintf( stderr, "NY is incorrect in file (got %d expected %d)\n",
	     nyin, ny );
if (gin != g) 
    fprintf( stderr, "G is incorrect in file (got %d expected %d)\n",
	     gin, g );

errcnt = 0;    
for (j=sz[1].start; j<=sz[1].end; j++) {
    int k;
    for (i=sz[0].start; i<=sz[0].end; i++) {
	k = i + sz[0].ndim * j;
	if (v[k] != v2[k]) {
	    errcnt++;
	    if (errcnt < 10) {
		fprintf( stderr, 
		   "[%d]Error in V values in file in location %d (%e != %e)\n",
		     PImytid, k, v[k], v2[k] );
		}
	    }
	}
    }


FREE( v );
FREE( v2 );
return 0;
}
