#ifndef lint
static char vcid[] = "$Id: pioreg.c,v 1.3 1995/02/13 16:43:20 gropp Exp $";
#endif

/*
     Routine to register a transport with the Parallel IO package.
*/
#include "comm/comm.h"
#include "system/nreg.h"
NRList *__PIOList = 0;

/*@
   PIORegister - Given a transport name (integer)  and a function pointer;
 adds the solver to the Parallel IO package.
 
   Input Parameters:
.      name  -  for instance PIOUnix, PIOUnitree.  If -1, a value will
                be chosen.
.      sname -  corresponding string for name
.      create - routine to create that transport
@*/
void PIORegister( name, sname, create )
int    name;
char   *sname;
void   (*create)();
{
if (!__PIOList) __PIOList = NRCreate();
if (name < 0) name = NRFindFreeId( __PIOList );
NRRegister( __PIOList, name, sname, (void *(*)())create );
}
/*----------------------------------------------------------------*/
/*@
   PIORegisterDestroy - Frees the list of SERIAL iterative solvers 
  which have been registered.
@*/
void PIORegisterDestroy()
{
if (__PIOList) {
    NRDestroy( __PIOList );
    __PIOList = 0;
    }
}

/*@
  PIOGetMethod - Given the argument list, return the selected method.

  Input parameters:
. Argc - pointer to arg count
. argv - argument vector
. flag - 1 if argument should be removed from list if found 
. sname - name used to indicate transport.  If null, -piomethod is used

  Output parameter:
. piomethod -  Parallel transport type
@*/
void PIOGetMethod( Argc, argv, flag, sname, piomethod )
int      *Argc, flag;
char     *argv, *sname;
int      *piomethod;
{
char sbuf[50];

if (!sname) sname = "-piomethod";
if (SYArgGetString( Argc, argv, flag, sname, sbuf, 50 )) {
    if (!__PIOList)
	PIORegisterAll();
    *piomethod = NRFindID( __PIOList, sbuf );
    }
}

/*@
    PIOIsMethodAvailable - Returns a method specified by name

    Input Parameter:
.   name - name of method, such as "ilu".

    Returns:
    Value of method or -1 if method is not available.  A method may not be
    available if (a) it is a vendor/installation specific method (b) methods
    have been removed by changing PIORegisterAll.
@*/
int PIOIsMethodAvailable( name )
char *name;
{
int meth;

if (!__PIOList) 
    PIORegisterAll();
return NRFindID( __PIOList, name );
}

#include <stdio.h>

/*@C
   PIOGetMethodName - Get the name (as a string) from the method type
 
   Input Parameter:
.  piometh - Parallel IO transport type
@*/
char *PIOGetMethodName( piometh )
int  piometh;
{
  if (!__PIOList) PIORegisterAll();
  return NRFindName( __PIOList, piometh );
}

#ifdef FOO
/*
   This routine takes the filename
*/
PIFileOps *PIOGetFileOpsFromName( name )
char *name;
{
PIFileOps *(*r)();

  /* Get the function pointers for the pio name requested */
  if (!__PIOList) {SVRegisterAll();} 
  if (!__PIOList) {
      SETERRC(1,"Could not acquire list of Parallel IO methods"); return 0;}
  r = (PIFileOps *(*)())NRFindRoutine( __PIOList, -1, name );
  if (!r) {SETERRC(1,"Unknown PIO method"); return 0;}
  return (*r)();
}
#endif
