#include "tools.h"
#include "system/system.h"
#include "comm/comm.h"
#include "comm/io/pio.h"
#include <math.h>

/* This is a test program.  It does multiple array writes 

   It is intended to test many of the options for performance and correctness.
   See the script runtest1
*/

main( argc, argv )
int  argc;
char **argv;
{
int worker();
PICall( worker, argc, argv );
}

int worker( argc, argv )
int  argc;
char **argv;
{
int         nx, ny, nr, nc, myid, ri, rj, i, j, g;
int         nxin, nyin, gin;
int         px, py;    /* Size of partition in each dimension */
PIFILE      *fp;
PIArrayPart sz[2];
int         nd;
int         errcnt;
double      *v, *v2, *v3, *vp;
double      t1, t2;
int         bytes;
int         rsize = 0, wsize = 0;
char        *fmat, *frmat;
int         flen;
int         err;
int         pmode;
int         nsks, nwf, nrf, avwlen, avrlen; /* Cache performance */
char        title[80];
char        fname[1024];
int         print_header = 0;

/* get the global size */
if (SYArgHasName( &argc, argv, 1, "-help" )) {
    fprintf( stderr, 
	    "%s -n nx -m ny -g n -rsize n -wsize n -block -pio -noformat\n", 
	     argv[0] );
    return 0;
    }
nx = 16;
ny = 16;
g  = 0;
SYArgGetInt( &argc, argv, 1, "-n", &nx );
SYArgGetInt( &argc, argv, 1, "-m", &ny );
SYArgGetInt( &argc, argv, 1, "-g", &g );
SYArgGetInt( &argc, argv, 1, "-rsize", &rsize );
SYArgGetInt( &argc, argv, 1, "-wsize", &wsize );
strcpy( fname, "foo" );
SYArgGetString( &argc, argv, 1, "-name", fname, 1024 );
title[0] = 0;
SYArgGetString( &argc, argv, 1, "-title", title, 80 );
if (SYArgHasName( &argc, argv, 1, "-block" ))
     PIFSetWritersBlock();
fmat  = "%12.4e\n";
frmat = "%12le\n";
flen  = 13;
if (SYArgHasName( &argc, argv, 1, "-noformat" )) {
    fmat  = 0;
    frmat = 0;
    flen  = sizeof(double);
    }
if (SYArgHasName( &argc, argv, 1, "-header" )) {
    if (PImytid == 0) {
	printf( 
" file size w MB/sec  seeks writes avelen r MB/sec seeks read avelen\n" );
	}
    }
pmode = PIO_AS_SEQUENTIAL;
if (SYArgHasName( &argc, argv, 1, "-pio" )) pmode = PIO_AS_PARALLEL;

fp = PIFopen( fname, PIAllProcs, O_WRONLY | O_CREAT, pmode, -1 );

if (rsize > 0 || wsize > 0) 
    PIFCacheEnable( fp, rsize, wsize );

PIWriteCommon( fp, "%4d",   4, &nx, 1, MSG_INT );
PIWriteCommon( fp, "%4d",   4, &ny, 1, MSG_INT );
PIWriteCommon( fp, "%4d\n", 5, &g,  1, MSG_INT );

nd   = 2;
nr   = PInumtids;
myid = PImytid;
if (nr > 1) 
    nr = sqrt( (double)nr + 0.1 );
nc = PInumtids / nr;
if (nr * nc != PInumtids) {
    fprintf( stderr, "Test only works for square arrays of processors\n" );
    return;
    }
ri = myid % nc;
rj = myid / nc;
    
px = nx / nc;
py = ny / nr;
/* printf( "[%d](%d,%d)\n", PImytid, ri, rj ); */
/* 
   The GLOBAL array is A(0:nx-1,0:ny-1).  The local array is
   declared as B(0:,0:), with B(g:,g:) matching A(ri*(nx/nc):,rj*(ny/nr):), 
   where the position of the processor in the mesh is given by (ri,rj)
 */
sz[0].mdim   = nx;
sz[0].ndim   = px + 2 * g;
sz[0].start  = g;
sz[0].end    = px - 1 + g;
sz[0].gstart = ri * px;
sz[0].gend   = sz[0].gstart + sz[0].ndim - 1;

sz[1].mdim   = ny;
sz[1].ndim   = py + 2 * g;
sz[1].start  = g;
sz[1].end    = py - 1 + g;
sz[1].gstart = rj * py;
sz[1].gend   = sz[1].gstart + sz[1].ndim - 1;
/* printf( "[%d]<%d,%d>\n", PImytid, sz[0].gstart, sz[1].gstart );  */
fflush(stdout);
v  = (double *)MALLOC( sz[0].ndim * sz[1].ndim * sizeof(double) );  CHKPTR(v);
vp = (double *)MALLOC( sz[0].ndim * sz[1].ndim * sizeof(double) );  CHKPTR(vp);
for (j=sz[1].start; j<=sz[1].end; j++) {
    for (i=sz[0].start; i<=sz[0].end; i++) {
	vp[i + sz[0].ndim * j] = 1.0e3 + 
	    i + sz[0].gstart + (j + sz[1].gstart) * sz[0].mdim;
	v[i + sz[0].ndim * j] = 
	    i + sz[0].gstart + (j + sz[1].gstart) * sz[0].mdim;
	}
    }

PIgsync( PIAllProcs );
t1 = SYGetElapsedTime();
PIWriteDistributedArray( fp, fmat, flen, sz, nd, v, MSG_DBL );
PIWriteDistributedArray( fp, fmat, flen, sz, nd, vp, MSG_DBL );
PIgsync( PIAllProcs );
t2 = SYGetElapsedTime() - t1;
if (PImytid == 0) {
    bytes = 2 * flen * nx * ny;
    if (title[0]) 
	printf( "# %s\n ", title );
    printf( "%9d %8.4f", bytes, bytes/(1.0e6*t2) );
    PIFCacheData( fp, &nsks, &nwf, &nrf, &avwlen, &avrlen );
    printf( " %4d %4d %6d", nsks, nwf, avwlen );
    fflush( stdout );
    }
PIFclose( fp );

/* Now, try to READ the data back; check that it is correct */
fp = PIFopen( fname, PIAllProcs, O_RDONLY, pmode, -1 );
if (!fp) {
    fprintf( stderr, "Can not open file for input\n" );
    return 0;
    }
if (rsize > 0 || wsize > 0) 
    PIFCacheEnable( fp, rsize, wsize );
v2 = (double *)MALLOC( sz[0].ndim * sz[1].ndim * sizeof(double) );  CHKPTR(v2);
v3 = (double *)MALLOC( sz[0].ndim * sz[1].ndim * sizeof(double) );  CHKPTR(v3);
for (i=0; i<sz[0].ndim * sz[1].ndim; i++) {
    v2[i] = -i;
    v3[i] = -i;
    }
PIReadCommon( fp, "%4d",   4, &nxin, 1, MSG_INT );
PIReadCommon( fp, "%4d",   4, &nyin, 1, MSG_INT );
PIReadCommon( fp, "%4d\n", 5, &gin,  1, MSG_INT );

PIgsync( PIAllProcs );
t1 = SYGetElapsedTime();
PIReadDistributedArray( fp, frmat, flen, sz, nd, v2, MSG_DBL );
PIReadDistributedArray( fp, frmat, flen, sz, nd, v3, MSG_DBL );
PIgsync( PIAllProcs );
t2 = SYGetElapsedTime() - t1;
if (PImytid == 0) {
    bytes = 2 * flen * nx * ny;
    printf( " %9d %8.4f ", bytes, bytes/(1.0e6*t2) );
    PIFCacheData( fp, &nsks, &nwf, &nrf, &avwlen, &avrlen );
    printf( "%4d %4d %6d\n", nsks, nrf, avrlen );
    fflush( stdout );
    }
PIFclose( fp );

/* Check that the values are correct */
if (nxin != nx) 
    fprintf( stderr, "NX is inccorect in file\n" );
if (nyin != ny) 
    fprintf( stderr, "NY is inccorect in file\n" );
if (gin != g) 
    fprintf( stderr, "G is inccorect in file\n" );

errcnt = 0;    
for (j=sz[1].start; j<=sz[1].end; j++) {
    int k;
    for (i=sz[0].start; i<=sz[0].end; i++) {
	k = i + sz[0].ndim * j;
	if (v[k] != v2[k]) {
	    errcnt++;
	    if (errcnt < 10) {
		fprintf( stderr, 
		   "[%d]Error in V values in file in location %d (%e != %e)\n",
		     PImytid, k, v[k], v2[k] );
		}
	    }
	if (vp[k] != v3[k]) {
	    errcnt++;
	    if (errcnt < 10) {
		fprintf( stderr, 
	   "[%d]Error in V2 values in file in location %d (%e != %e)\n",
	     PImytid, k, vp[k], v3[k] );
		}
	    }
	}
    }

if (errcnt > 0) {
    fprintf( stderr, "[%d] Total number of errors = %d\n", PImytid, errcnt );
    }
PIgisum( &errcnt, 1, &i, PSAllProcs );

FREE( v );
FREE( v2 );
FREE( vp );
FREE( v3 );
return errcnt;
}
