#ifndef lint
static char vcid[] = "$Id: doctext.c,v 1.6 1995/02/13 16:54:36 gropp Exp $";
#endif

#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "tools.h"
#include "doc/doc.h"

/* Set this to 0 for LaTeX output */
static int IsMan = 1;
static int IsLaTeX = 0;

static int DebugDoc = 0;

/* Pointer to output function structure */
static DocOutput *outfcn = 0;

/* Indicate whether the "location" field should be written.  For example,
   a location of a "readme" file is often not useful
 */
static int GiveLocation = 1;

/* This is used to decide whether '' and `` change fonts */
int DoQuoteFormat = 0;

/* This is designed to work with comments in C programs.  Comments to
   be processed by this routine have the form:
    / * <char> .... <char> * /
   where <char> is one of the characters in MATCH_CHARS.  The definitions
   of the characters follows
 */

/*
    The next change is to attach to each "action" a routine that is called;
    then this code can be shared with the c2fort, lint, and info code.
 */ 

/*D
doctext - program to extract man pages from source files

Input Arguments:
$ -mpath path
Sets the path where the man pages will be written
$ -ext n
Sets the extension (1-9,l, etc)
$ -I filename
filename contains the public includes needed by these routines; it will
automatically be added to the synopsis of the routines
$ -latex
Generate latex output rather than man (nroff) format files.
$ -html
Generate html (WWW) output rather than man (nroff) format files.
$ -index filename
Generate a index file appropriate for tohtml (for generating WWW files)
This APPENDS to any existing file; make sure that you don't add duplicate
entries.
$ -indexdir dirname
Sets the root directory for the index file entries.  This allows you to 
specify an absolute URL for the generated files to be used in the generated
index.  For example, 
$ -index foo.cit -indexdir "http:/www.mcs.anl.gov/home/gropp/man"
$
$ -jumpfile filename
$ Generate a jump file appropriate for emacs-19 and the command 
lookup-petsc-routine-from-text.  
$
$ -outfile filename
Put the manpages in the indicated file.  Normally, a separate file is
created for each routine.  This may be appropriate for html or win output
(win not yet supported).
$ -heading name 
Name for the heading (middle of top line)
$filenames
Names of the files from which documents are to be extracted

Notes:
The output file names are formed from a path, the name of the routine, and
an extension.  The format of the output is a man page, suitable for use
by man, nroff -man, or xman.

Typesetting Conventions:
A $<space> where $ is in column 1 turns into a new paragraph.  This allows the
insertion of line breaks into the man page (ordinarily, leading blanks
are deleted as is needed by n/troff).

A .<space> in the first column is used to indicate a routine argument; these
are set in a special way.  

There are some further restrictions if a "whatis" file is to be correctly
generated (this is necessary for man -k to work).  In this case, the first
few lines MUST look like

$    name - short description

$    sometext:

where sometext ends a line with a ":".  

Author: 
Bill Gropp
D*/
/* In the past, we said:

   Other .<non-space> commands are passed through, 
   allowing you to use any of the nroff commands.

   But I think that this is a bad idea.  I'd like to formalize most of
   the fields, and give them special processing (particularly for the
   TeX version)

   A previous version of this had statically allocated fields for the
   output filename, routine name, path, and includefile; these cause problems
   for PC versions that have to operate with a small stack for local
   variables.
 */

/* Form the filename.  Handles making SHORTER names */
void MakeFileName( path, routine, ext, outfilename )
char *path, *routine, *ext, *outfilename;
{
#ifndef __MSDOS__
sprintf( outfilename, "%s%s.%s", path, routine, ext );
#else
char *p;
strcpy( outfilename, path );
p = outfilename;
while (*p) { if (*p == '/') *p = '\\'; p++; }
strcat( outfilename, "\\" );
/* Eventually, this could be "remove vowels" or something like that */
strncat( outfilename, routine, 8 );
strcat( outfilename, "." );
strcat( outfilename, ext );
#endif
}
 
int main( argc, argv )
int  argc;
char **argv;
{
char *outfilename;
char *routine;
char *path, extension[10];
char lextension[10];
char date[50];
char *heading;
char *incfile, userlevel[10];
char *indexfile, *infilename, *basedir, *baseoutfile, *jumpfile;
FILE *fd, *fout, *incfd, *idxfd, *jumpfd;
char *idxdir;
int  nread;
char kind;

outfilename = (char *)MALLOC( MAX_FILE_SIZE );    CHKPTRN(outfilename);
path        = (char *)MALLOC( MAX_FILE_SIZE );    CHKPTRN(path);
incfile     = (char *)MALLOC( MAX_FILE_SIZE );    CHKPTRN(incfile);
indexfile   = (char *)MALLOC( MAX_FILE_SIZE );    CHKPTRN(indexfile);
jumpfile    = (char *)MALLOC( MAX_FILE_SIZE );    CHKPTRN(jumpfile);
basedir     = (char *)MALLOC( MAX_FILE_SIZE );    CHKPTRN(basedir);
baseoutfile = (char *)MALLOC( MAX_FILE_SIZE );    CHKPTRN(baseoutfile);
idxdir      = (char *)MALLOC( MAX_FILE_SIZE );    CHKPTRN(idxdir);
routine     = (char *)MALLOC( MAX_ROUTINE_NAME ); CHKPTRN(routine);
heading     = (char *)MALLOC( MAX_ROUTINE_NAME ); CHKPTRN(heading);

/* set the defaults */
strcpy( path, "./" );
strcpy( extension, "3" );
incfile[0]     = '\0';
indexfile[0]   = '\0';
jumpfile[0]    = '\0';
strcpy( idxdir, "." );
userlevel[0]   = '\0';
basedir[0]     = '\0';
baseoutfile[0] = '\0';
strcpy( heading, "PETSc" );

/* check for option -extension and -mpath */
nread = HandleArgs( argc, argv, path, extension, incfile,
		    userlevel, indexfile, idxdir, jumpfile, 
		    basedir, baseoutfile, heading );
argc -= nread;
argv += nread;

if (!outfcn)
    outfcn = CreateOutputMan();
    
/* Open up the file of public includes */    
if (incfile[0])
    incfd = fopen( incfile, "r" );
else
    incfd = 0;

/* Open up the index file */    
if (indexfile[0])
    idxfd = fopen( indexfile, "a" );
else
    idxfd = 0;

/* Open up the jump file */    
if (jumpfile[0])
    jumpfd = fopen( jumpfile, "a" );
else
    jumpfd = 0;

/* Open user-selected output file */
if (baseoutfile[0]) {
    strcpy( outfilename, baseoutfile );
    fout = fopen( outfilename, "w" );
    if (!fout) {
        fprintf( stderr, "Could not open output file %s\n", outfilename );
        return 1;
        }
    if (basedir[0]) OutputBOF( fout, basedir );        
    }

/* process all of the files */
while (argc--) {
    infilename = *argv++;
    if (DebugDoc) printf( "About to open %s\n", infilename );
    fd = fopen( infilename, "r" );
    if (!fd) {
    	fprintf( stderr, "Could not open file %s\n", infilename );
    	continue;
        }
    ResetLineNo();
    LastChangeToFile( infilename, date ); 
    /* At this point, we can trim the filename of any leading trash, 
       such as "./".  Later.... */
    while (FoundLeader( fd, routine, &kind )) {
	/* At this point, we could modify the routine name to make it
	   shorter */
	if (kind == INCLUDE) {
	    /* Eventually, cache the includes for this file */
	    continue;
	    }
	strcpy( lextension, extension );
	if (kind == FORTRAN) 
	    strcat( lextension, "f" );
	if (kind == TEXT) {
	    /* Save a block of common text */
	    SaveName( fd, routine );
	    continue;
	    }
	if (!baseoutfile[0]) {
	    MakeFileName( path, routine, lextension, outfilename );
    	    fout = fopen( outfilename, "w" );
    	    if (!fout) {
    	        fprintf( stderr, 
			"Could not open output file %s\n", outfilename );
    	        break;
    	        }
    	    }
    	if (idxfd) {
	    /* If there is an indexdir, we need to remove all directories
	       from the filename */
	    char baseoutfile[1024];
	    char *pp;
	    pp = outfilename + strlen(outfilename) - 1;
	    while (pp > outfilename && pp[0] != '/') pp--;
	    pp++;
    	    fprintf( idxfd, "man:+%s++%s++++man+%s/%s#%s\n", 
    	             routine, routine, idxdir, pp, routine );
    	    }
	if (jumpfd) {
	    char rpath[1024];
	    /* If there is a jumpfile, add the line */
	    /* Note that we want an ABSOLUTE path for the infilename */
	    SYGetRealpath( infilename, rpath );
	    fprintf( jumpfd, "%s:%s:%d\n", routine, rpath, GetLineNo() );
	    }
        if (!baseoutfile[0] && basedir[0]) OutputBOF( fout, basedir );        
    	OutputManPage( fd, fout, routine, lextension, infilename, kind, date,
		       incfd, userlevel, heading );
        if (!baseoutfile[0]) fclose( fout );
	}
    fclose( fd );
    }
if (baseoutfile[0]) fclose( fout );
return 0;
}

/* 
   There are a number of things to watch for.  One is that leading blanks are
   considered significant; since the text is being formated, we usually don't
   agree with that. 

   userlevel is a refugee from some time; it is not used.
 */
OutputManPage( fin, fout, name, level, filename, kind, date, 
	       incfd, userlevel, heading )
FILE *fin, *fout;
char *name, *level, *filename, kind, *date, *userlevel, *heading;
FILE *incfd;
{
int  i, c, nsp;
char token[1024];
int  has_synopsis;

has_synopsis = 0;
OutputText( fin, fout, name, level, filename, kind, &has_synopsis, 
	    date, heading );

if (kind == ROUTINE && !has_synopsis) {
    /* This is a routine, so we can extract the exact usage */    
    /* Now, hunt for the first open brace.  Output with "obeylines".
       Note that we copy Synposis; some systems put strings into
       read-only pages */
    strcpy( token, "Synopsis" );
    OutputSection( fout, token );
    OutputVerbatimBegin( fout );
    /* include the includes that the user needs */
    if (incfd) {
	fseek( incfd, 0L, 0 );
	while ((c = DocGetChar( incfd )) != EOF) 
	    OutputChar( fout, (char)c );
	}
    /* We may also need to add some includes private to a file.  I haven't
       decided how to do that yet */

    SkipWhite( fin );
    while (1) {
	c = FindNextToken( fin, token, &nsp );
	if (c == EOF || token[0] == '{') break;
	/* Don't output "register" */
	if (strcmp( token, "register" ) != 0) {
	    for (i=0; i<nsp; i++) OutputChar( fout, ' ' );
	    OutputString( fout, token );
	    }
	}
    if (IsMan)
	OutputString( fout, "\n" );
    OutputVerbatimEnd( fout );
    }
    
/* Now add the filename where the routine or description is located */
if (filename && GiveLocation) {
    /* Optionally, add the directory? */
    OutputLocation( fout, filename );
    }
OutputEndPage( fout );
}

/*
    This routine generates the text part of a text page.
    It has to do several things:
    Look for the end of the text (\n<white><kind> * /<white>\n)
    Process formating directives
        a: \n<white><string>:\n
        b: \n.<string>\n
        c: \n$<string>\n
    So the rule is:
    After seeing a newline, check for . and $
    if neither, skip white space.
    Copy rest of line into buffer.
    check for <kind> * / or :\n.
    Output line accordingly.
 */        
OutputText( fin, fout, name, level, filename, kind, has_synopsis, date, 
	    heading )
FILE *fin, *fout;
char *name, *level, *filename;
char kind;
int  *has_synopsis;
char *date, *heading;
{
int  c;
char lineBuffer[MAX_LINE], *lp;
int  firsthyphen;
int  lastWasNl;
int  doing_synopsis = 0;
	
firsthyphen = 1;
/* After the name, if there IS a -, prefix it with a \.  This seems to be
   necessary for making the whatis file */
lastWasNl   = 1;
/* Note that the NAME field can't use a font changes */
OutputTitle( fout, name, level, date, heading );
lineBuffer[0] = '+';   /* Sentinal on lineBuffer */
while (1) {
    lp = lineBuffer + 1;
    c  = DocGetChar( fin );
    if (c == EOF) break;
    if (doing_synopsis && (c == VERBATIM || c == ARGUMENT || c == '\n')) {
	OutputVerbatimEnd( fout );
	doing_synopsis = 0;
	}
    if (c == VERBATIM) {
	/* Raw mode output. */
	ProcessVerbatimFmt( fin, fout, outfcn, &lastWasNl );
        }
    else if (c == ARGUMENT) {
	ProcessDotFmt( fin, fout, outfcn, &lastWasNl );
        }
    else if (c == '\n') { 
	if (lastWasNl) OutputEndPar( fout );
	else           OutputChar( fout, '\n' );
	lastWasNl = 1;
        }
    else {
	if (isspace(c) && c != '\n')
	    SkipWhite( fin );
	else {
            if (firsthyphen && c == '-') {
		if (IsMan) 
		    OutputRawChar( fout, '\\' );  /* *lp++ */
		else if (IsLaTeX) {
		    /* Turn it into an em-dash */
		    OutputChar( fout, c );
		    OutputChar( fout, c );
		    }
		}
            firsthyphen = 0;
	    *lp++ = c;
            }
    	/* Copy to end of line; do NOT include the EOL */
    	while ((c = DocGetChar( fin )) != EOF && c != '\n') 
    	    *lp++ = c;
    	lp--;
    	while (isspace(*lp)) lp--;
    	lp[1] = '\0';    /* Add the trailing null */
    	if (lineBuffer[1] == kind && strcmp(lineBuffer+2,"*/") == 0)
    	    break;
    	else if (lp[0] == ':') {
    	    *lp = '\0';
	    /* Using .SS confused the whatis builder */
	    OutputSection( fout, lineBuffer + 1 );
	    /* If there is a "manual" SYNOPSIS, don't generate one later */
	    if (MatchTokens( lineBuffer+1, "SYNOPSIS" ) == 0) {
		*has_synopsis = 1;
		doing_synopsis= 1;
		OutputVerbatimBegin( fout );
		/* Need to end the synopsis with .fi */
		}
    	    }
    	else {
	    OutputString( fout, lineBuffer + 1 );
	    OutputString( fout, "\n" );
    	    }
        }
    }
if (doing_synopsis) 
    OutputVerbatimEnd( fout );
}

/* Routine to process the arguments.  Returns the number read */
int HandleArgs( argc, argv, path, extension,
                incfile, userlevel, indexfile, idxdir, jumpfile, 
	        basedir, baseoutfile, heading )
int  argc;
char **argv;
char *path, *extension, *incfile, *userlevel, *indexfile, *jumpfile, *basedir,
     *baseoutfile, *idxdir, *heading;
{
int nread = 0, ln;
argc--; argv++; nread++; /* skip the program */

/* Look first for -help or -h */
if (SYArgHasName( &argc, argv, 1, "-help" ) ||
    SYArgHasName( &argc, argv, 1, "-h" )) {
    PrintHelp( );
    exit( 0 );
    }
nread += 2*SYArgGetString( &argc, argv, 0, "-mpath", path, 1024 );
ln    = strlen(path);
if (path[ln-1] != '/') {
    path[ln] = '/'; 
    path[ln+1] = '\0';
    }
nread += 2*SYArgGetString( &argc, argv, 0, "-ext",     extension, 3 );
nread += 2*SYArgGetString( &argc, argv, 0, "-I",       incfile,  1024 );
nread += 2*SYArgGetString( &argc, argv, 0, "-index",   indexfile,  1024 );
nread += 2*SYArgGetString( &argc, argv, 0, "-indexdir",idxdir, 1024 );
nread += 2*SYArgGetString( &argc, argv, 0, "-jumpfile",jumpfile,  1024 );
nread += 2*SYArgGetString( &argc, argv, 0, "-basedir", basedir,  1024 );
nread += 2*SYArgGetString( &argc, argv, 0, "-outfile", baseoutfile,  1024 );
nread += 2*SYArgGetString( &argc, argv, 0, "-l",       userlevel, 10 );
nread += 2*SYArgGetString( &argc, argv, 0, "-heading", heading, 10 );
if (SYArgHasName( &argc, argv, 0, "-latex" )) {
    nread += 1;
    outfcn = CreateOutputLaTeX( );
    IsMan   = 0;
    IsLaTeX = 1;
    }
if (SYArgHasName( &argc, argv, 0, "-html" )) {
    nread += 1;
    outfcn = CreateOutputHTML( );
    strcpy( extension, "html" );
    IsMan = 0;
    }
if (SYArgHasName( &argc, argv, 0, "-debug" )) {
    nread++;
    DebugDoc = 1;
    }
if (SYArgHasName( &argc, argv, 0, "-nolocation" )) {
    nread++;
    GiveLocation = 0;
    }
if (SYArgHasName( &argc, argv, 0, "-quotefmt" )) {
    nread++;
    DoQuoteFormat = 1;
    }    
return nread;
}

/* Print the Help page for doctext */
PrintHelp()
{
fprintf( stderr, "\
doctext [ -mpath path ] [ -ext n ] [ -I filename ] [ -latex ]\n\
        [ -html ] [ -index filename ] [ -indexdir filename ]\n\
        [ -jumpfile filename ] [ -outfile filename ] \n\
        [ -heading string ] [ -basedir dirname ] filenames\n\
\n\
    -mpath path    Sets the path where the man pages will be written\n\
    -ext n         Sets the extension (1-9,l, etc)\n\
    -nolocation    Don't give the filename where the man page info was\n\
    -I filename    Filename contains the public includes needed by these\n\
                   routines\n\
    -latex         Generate latex output rather than man format files\n\
    -html          Generate html (WWW) output rather than man format files\n\
    -index filename Generate a index file appropriate for tohtml\n\
                    (for generating WWW files)\n" );
fprintf( stderr, "\
    -indexdir dirname \n\
                   Sets the root directory for the index file entries.\n\
                   For example\n\
          -index foo.cit -indexdir \"http://www.mcs.anl.gov/home/gropp/man\"\n\
    -outfile filename\n\
                   Put the manpages in the indicated file. \n\
    -heading name  Name for the heading (middle of top line)\n\
    -quotefmt      support '\\tt' and `\\em`\n\
    filenames      Names of the files from which documents are to be\n\
                   extracted\n\
\n\
See the man page for details\n" );
}

/*
  man macros that are available:

  .TH t s c n 
   Title t, Section s, commentary c (e.g., local), and manual name n

  .SH text
   Subhead text

  .SS text
   Sub-Subhead text
 
  .B text
   Bold text

  .I text
   Italics text

  .SM text
   Make text 1 point smaller

  .P 
   Begin paragraph

  .BP in
    Begin paragraph with hanging indent

  .PD v
    Set interparagraph spacing to v vertical spaces
 */

