#ifndef lint
static char vcid[] = "$Id: outhtml.c,v 1.3 1995/02/13 16:54:49 gropp Exp $";
#endif

#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include "tools.h"
#include "doc/doc.h"

/*
    This file contains routines to generate html output for the structured
    documentation commands
 */

static int InVerbatim = 0; /* ????? */
static int LastWasEndOfDescription = 0;

/* These are used to implement '\tt' and `\em`  */
extern int DoQuoteFormat;
static int EmCnt = 1;
static int TtCnt = 1;

/* Beginning of file output */
void OutputBOFHTML( fout, basedir )
FILE *fout;
char *basedir;
{
/* The name should look something
   like "http://www.mcs.anl.gov/home/gropp/index.html" 
*/
fprintf( fout, "<base href=\"%s\">\n", basedir );	
}

void OutputCharHTML( fout, c )
FILE *fout;
char c;
{
    if (!InVerbatim) {
	if (c == '<')
	    fputs( "&lt;", fout );
	else if (c == '>') 
	    fputs( "&gt;", fout );
	else if (c == '&')
	    fputs( "&and;", fout );
        else if (c == '`' && DoQuoteFormat) {
    	    /* Emphasis is `  ` */
    	    if (EmCnt & 0x1) fputs( "<em>", fout );
    	    else             fputs( "</em>", fout );
    	    EmCnt++;
            }
        else if (c == '\'' && DoQuoteFormat) {
    	    /* Tt is ' ' */
    	    if (TtCnt & 0x1) fputs( "<tt>", fout );
    	    else             fputs( "</tt>", fout );
    	    TtCnt++;
	    }
	else
	    putc( c, fout );
	}
    else
	putc( c, fout );
}

void OutputRawCharHTML( fout, c )
FILE *fout;
char c;
{
putc( c, fout );
}

void OutputStringHTML( fout, s )
FILE *fout;
char *s;
{
char *p, c, *rstr = 0;

/* look for special characters */
p = s;
while (*p) {
    c    = *p;
    rstr = 0;
    if (!InVerbatim) {
	if (c == '<')
	    rstr = "&lt;";
	else if (c == '>') 
	    rstr = "&gt;";
	else if (c == '&')
	    rstr = "&amp;";
        else if (c == '`' && DoQuoteFormat) {
    	    /* Emphasis is `  ` */
    	    if (EmCnt & 0x1) rstr = "<em>";
    	    else             rstr = "</em>";
    	    EmCnt++;
            }
        else if (c == '\'' && DoQuoteFormat) {
    	    /* Tt is ' ' */
    	    if (TtCnt & 0x1) rstr = "<tt>";
    	    else             rstr = "</tt>";
    	    TtCnt++;
	    }
	}

    if (rstr) {
        *p = '\0';
        fputs( s, fout );
        fputs( rstr, fout );
	s = p + 1;
        }
    p++;
    }
if (*s) 
    fputs( s, fout );
}

void OutputBlankHTML( fout )
FILE *fout;
{
fputs( " ", fout );
}

void OutputLocationHTML( fout, s )
FILE *fout;
char *s;
{
fprintf( fout, "<P><b>Location:</b> %s<p>\n", s );
}

void OutputTitleHTML( fout, name, level, date, heading )
FILE *fout;
char *name, *level, *date, *heading;
{
char bname[256];
strcpy( bname, name );
/* We have to be prepared to correct the name if it contains a
   reserved character */
fprintf( fout, "<title>%s</title>\n", name );
fprintf( fout, "<a name=\"%s\"><h1>%s</h1></a>\n", name, name );   
}

void OutputSectionHTML( fout, name )
FILE *fout;
char *name;
{
fprintf( fout, "<h2>%s</h2>\n", name );
}

void OutputLineBreakHTML( lastnl, fout )
int  lastnl;
FILE *fout;
{
fputs( "<BR>\n", fout );
}

void OutputArgBeginHTML( fout )
FILE *fout;
{
if (!LastWasEndOfDescription)
    fputs( "<dl>", fout );
fputs( "<dt>", fout );
fputs( "<b>", fout );
}

void OutputArgDefnHTML( fin, fout )
FILE *fin, *fout;
{
int c;

SkipWhite( fin );
fputs( "</b>", fout );
fputs( "<dd>", fout );
c = getc( fin );
if (c != '-') 
    ungetc( (char)c, fin );
else
    SkipWhite( fin );
}

void OutputArgEndHTML( fout )
FILE *fout;
{
/* LastWasEndOfDescription = 1; */
fputs( "</dl>\n", fout );
}

void OutputVerbatimBeginHTML( fout )
FILE *fout;
{
LastWasEndOfDescription = 0;
fputs( "<pre><code>\n", fout );
InVerbatim = 1;
}

void OutputVerbatimEndHTML( fout )
FILE *fout;
{
fputs( "</code></pre>", fout );
InVerbatim = 0;
}

void OutputEndPageHTML( fout )
FILE *fout;
{
LastWasEndOfDescription = 0;
}

void OutputEndParHTML( fout )
FILE *fout;
{
fputs( "<P>\n", fout );	
}

void OutputStartEmHTML( fout )
FILE *fout;
{
fputs( "<em>", fout );
}

void OutputEndEmHTML( fout )
FILE *fout;
{
fputs( "</em>", fout );
}

void OutputStartTtHTML( fout )
FILE *fout;
{
fputs( "<tt>", fout );
}

void OutputEndTtHTML( fout )
FILE *fout;
{
fputs( "</tt>", fout );
}	

void OutputPictureHTML( fout, name )
FILE *fout;
char *name;
{
char gifname[1024], *p;
int  ln;

/* Get gif version of filename */
ln = strlen(name);
if (ln > 1020) {
    fprintf( stderr, "Name %s is too long!\n", name );
    return;
    }
strcpy( gifname, name );
p = gifname + ln - 1;
while (p > gifname && *p != '.') p--;
strcpy( p + 1, "gif" );

/* If file does not exist, then create it (if can't create, generate button?)
 */

fprintf( fout, "<IMG \"%s\">\n", gifname );
}

void OutputStartCaptionHTML( fout )
FILE *fout;
{
/* May want to use blockquote ? */	
}

void OutputEndCaptionHTML( fout )
FILE *fout;
{
fputs( "<P>\n", fout );
}

DocOutput *CreateOutputHTML( )
{
DocOutput *new;

new = NEW( DocOutput );
new->outbof           = OutputBOFHTML;
new->outchar          = OutputCharHTML;
new->outraw           = OutputRawCharHTML;
new->outstring        = OutputStringHTML;
new->outblank         = OutputBlankHTML;
new->outlocation      = OutputLocationHTML;
new->outtitle         = OutputTitleHTML;
new->outsection       = OutputSectionHTML;
new->outlinebreak     = OutputLineBreakHTML;
new->outargbegin      = OutputArgBeginHTML;
new->outargdefn       = OutputArgDefnHTML;
new->outargend        = OutputArgEndHTML;
new->outverbatimbegin = OutputVerbatimBeginHTML;
new->outverbatimend   = OutputVerbatimEndHTML;
new->outendpage       = OutputEndPageHTML;
new->outendpar        = OutputEndParHTML;
new->outstartem       = OutputStartEmHTML;
new->outendem         = OutputEndEmHTML;
new->outstarttt       = OutputStartTtHTML;
new->outendtt         = OutputEndTtHTML;
new->outpicture       = OutputPictureHTML;
new->outstartcaption  = OutputStartCaptionHTML;
new->outendcaption    = OutputEndCaptionHTML;

return new;
}
