
#ifndef __KSP
#define __KSP
#include "packages.h"

#if !defined(VEC_COMPONENT)
  Vector packages must be installed for KSP package
#endif

#include "vec.h"

typedef struct _KSPCtx*     KSPCtx;

/*  Possible Krylov Space Methods */
typedef enum { KSPRICHARDSON, KSPCHEBYCHEV, KSPCG, KSPGMRES, 
               KSPTCQMR, KSPBCGS, KSPCGS, KSPTFQMR, KSPCR, KSPLSQR,
               KSPPREONLY } KSPMETHOD;

/* Routines */
#ifdef ANSI_ARG
#undef ANSI_ARG
#endif
#ifdef __STDC__
#define ANSI_ARGS(a) a
#else
#define ANSI_ARGS(a) ()
#endif

extern KSPCtx   KSPCreate          ANSI_ARGS(());
extern void     KSPSetMethod       ANSI_ARGS((KSPCtx,KSPMETHOD));
extern void     KSPSetUp           ANSI_ARGS((KSPCtx));
extern int      KSPSolve           ANSI_ARGS((KSPCtx));
extern void     KSPDestroy         ANSI_ARGS((KSPCtx));

extern void     KSPRegisterAll     ANSI_ARGS(());
extern void     KSPRegister        ANSI_ARGS((KSPMETHOD,char *,void (*)()));

extern char     *KSPGetMethodName  ANSI_ARGS((KSPMETHOD));
extern void     KSPGetWorkCounts 
                           ANSI_ARGS((KSPCtx,int*,int*,int*,int*,int*));
extern void     KSPClearWorkCounts ANSI_ARGS((KSPCtx));

extern void     KSPSetIterations   ANSI_ARGS((KSPCtx,int));
extern void     KSPSetRightPreconditioner ANSI_ARGS((KSPCtx));
extern int      KSPGetPreconditionerSide ANSI_ARGS((KSPCtx));
extern KSPMETHOD  KSPGetMethodFromContext  ANSI_ARGS((KSPCtx));
extern int      KSPGetMethodFromCommandLine
                    ANSI_ARGS((int *,char **,int,char *,KSPMETHOD *));
extern void     KSPSetRelativeTolerance  ANSI_ARGS((KSPCtx,double));
extern void     KSPSetAbsoluteTolerance ANSI_ARGS((KSPCtx,double));
extern void     KSPSetDivergenceTolerance  ANSI_ARGS((KSPCtx,double));
extern void     KSPSetCalculateResidual ANSI_ARGS((KSPCtx));
extern void     KSPSetDoNotCalculateResidual ANSI_ARGS((KSPCtx));
extern void     KSPSetUsePreconditionedResidual ANSI_ARGS((KSPCtx));
extern void     KSPSetInitialGuessZero ANSI_ARGS((KSPCtx));
extern void     KSPSetCalculateEigenvalues ANSI_ARGS((KSPCtx));
extern void     KSPSetRhs ANSI_ARGS((KSPCtx,VeVector));
extern VeVector KSPGetRhs ANSI_ARGS((KSPCtx));
extern void     KSPSetSolution ANSI_ARGS((KSPCtx,VeVector));
extern VeVector KSPGetSolution ANSI_ARGS((KSPCtx));
extern void     KSPSetAmult  
               ANSI_ARGS((KSPCtx,void (*)(void *,VeVector,VeVector),void *));

extern void     *KSPGetAmultContext ANSI_ARGS((KSPCtx));
extern void     KSPSetAmultTranspose 
                       ANSI_ARGS((KSPCtx,void (*)(void *,VeVector,VeVector)));
extern void     KSPSetBinv  
               ANSI_ARGS((KSPCtx,void(*)(void *,VeVector,VeVector),void *));
extern void     *KSPGetBinvContext ANSI_ARGS((KSPCtx));
extern void     KSPSetBinvTranspose  
                    ANSI_ARGS((KSPCtx,void (*)(void *,VeVector,VeVector)));
extern void     KSPSetMatop 
                    ANSI_ARGS((KSPCtx,void (*)(void *,VeVector,VeVector)));
extern void     KSPSetMatopTranspose  
                    ANSI_ARGS((KSPCtx,void (*)(void *,VeVector,VeVector)));
extern void     KSPSetMonitor 
               ANSI_ARGS((KSPCtx,void (*)(KSPCtx,int,double, void*), void *));
extern void     *KSPGetMonitorContext  ANSI_ARGS((KSPCtx));
extern void     KSPSetResidualHistory  ANSI_ARGS((KSPCtx, double *,int));
extern void     KSPSetConvergenceTest  
               ANSI_ARGS((KSPCtx,int (*)(KSPCtx,int,double, void*), void *));
extern void     *KSPGetConvergenceContext  ANSI_ARGS((KSPCtx));

extern VeVector KSPBuildSolution  ANSI_ARGS((KSPCtx, VeVector));
extern VeVector KSPBuildResidual  ANSI_ARGS((KSPCtx, VeVector, VeVector));

extern void     KSPRichardsonSetScale  ANSI_ARGS((KSPCtx , double));
extern void     KSPChebychevSetEigenvalues  
                                  ANSI_ARGS((KSPCtx , double, double));
extern void     KSPGMRESSetDirections  ANSI_ARGS((KSPCtx, int));

extern void     KSPDefaultCGMonitor  ANSI_ARGS((KSPCtx,int,double, void * ));
extern int      KSPDefaultCGConverged  ANSI_ARGS((KSPCtx,int,double, void *));
extern void     KSPDefaultMonitor    ANSI_ARGS((KSPCtx,int,double, void *));
extern void     KSPLineGraphMonitor  ANSI_ARGS((KSPCtx,int,double, void *));
extern int      KSPDefaultConverged  ANSI_ARGS((KSPCtx,int,double, void *));

#endif


