.TH Linking r "1/3/1995" " " "PETSc"
.SH NAME
Linking \- Linking with the PETSc libraries

.SH LIBRARY ORGANIZATION

The object libraries for the PETSc package are organized by
level of optimization and architecture as

.br
  libs/libs[g,O,Opg]/<architecture>/*.a

For example, the libraries for production (optimized) runs for the
sun4 are in

.br
  libs/libsO/sun4/*.a

The PETSc libraries are subdivided by components, as listed below,
roughly in increasing levels of abstraction.
.br
.br
  libpecore.a     - system portability
.br
  libpetsc.a      - vectors and miscellaneous routines
.br
  libpemat.a      - matrices
.br
  libpetscgrf.a   - graphics
.br
  libiter.a       - Krylov space methods
.br
  libsles.a       - linear equations solvers
.br
  libsnes.a       - nonlinear equations solvers
.br
  domain.a        - PDE solvers
.br
  libpetscfort.a  - Fortran interface

The parallel libraries also allow specification of the type of
communication library, such as MPI, EUI, P4, PVM, etc.  As shown
below, this is indicated by <comm>, where <comm> = mpi, eui, p4,
pvm, etc.  Note that <comm> need not be specified for certain
architectures such as the Intel DELTA, when the vendor-supplied
communication library is used.
.br
.br
  libpecomm<comm>.a    - miscellaneouos parallel 
.br
     routines, including Chameleon 
.br
  libpeblkcm<comm>.a   - BlockComm routines 
.br
  libpetscgrf<comm>.a  - parallel graphics
.br
  libpetscfort<comm>.a - parallel Fortran interface

The libraries with which users must link depend upon the component(s)
of PETSc being used.  The examples directory within each PETSc
component contains a Makefile that specifies the appropriately ordered
library list for that particular component.  The order in which the
libraries are linked (from higher level modules to lower level modules)
is critically important for successful linking.

.SH UNIPROCESSOR LIBRARIES

The variable $(LDIR), as defined within the Makefiles by the configure
installation process, indicates the PETSc library directory for a
particular architecture and level of optimization.  The following
sort of statement, as found within solvers/examples/Makefile, specifies
the libraries for using the uniprocessor SLES (linear equations
solvers) package
.br
.br
  LIBS = -L$(LDIR) -lsles -liter -lpetscgrf -lpemat \\
.br
         -lpetsc -lpecore $(BLAS) $(X11LIB) $(VLIBS) \\
.br
         $(FCLIB) -lm $(SLIB)

The above variables are defined during the configure process as follows
.br
.br
  BLAS   - BLAS library
.br
  SLIB   - any required system libraries
.br
  X11LIB - X11 window system library (optional, used with libpetscgrf.a ) 
.br
  FCLIB  - libraries for a C linker to use code containing Fortran code

See the man page ToolsMake for additional information.

.SH MULTIPROCESSOR LIBRARIES

When working in a parallel message-passing environment
libpecomm<comm>.a is also necessary, where <comm> optionally
indicates an alternative type of communication library.  For
example, comm/examples/Makefile specifies the libraries for
using the Chameleon message-passing interface,
.br
.br
  LIBS = -L$(LDIR) -lpecomm<comm> -lpetsc -lpecore \\
.br
         -lpecomm<comm> $(CLIB) $(SLIB) -lm
.br
The variable $(CLIB), as defined by the GNU configure installation
process, specifies the required communication libraries (such as
libp4.a or libpvm.a).

A conflict can arise when using libpecomm<comm>.a and the
uniprocessor PETSc libraries, since most of the routines in
libpecomm<comm>.a have single-processor counterparts.  Parallel
programs that seem to hang may be doing so as a result of linking
with the uniprocessor modules instead of libpecomm<comm>.a .
If, for some reason, libpecomm<comm>.a was not built correctly,
the same problem can occur.

.SH PETSC FORTRAN INTERFACE LIBRARIES

When using Fortran to call the PETSc routines, additional libraries
are required.  These contain routines that translate from Fortran
to the PETSc libraries, and must be placed before them on the link
line.  The Fortran interface libraries are

.br
  libpetscfort<comm>.a - parallel Fortran interface
.br
  libpetscfort.a - Fortran interface

As above, the "comm" version is required only when using the parallel
routines.  For example, the complete set of libraries needed to use
a Fortran program to call routines from uniprocessor SLES is
.br
.br
  LIBS = -L$(LDIR) -lpetscfort -lsles -liter -lpetscgrf \\
.br
         -lpemat -lpetsc -lpecore $(BLAS) $(X11LIB) \\
.br
         $(VLIBS) $(FCLIB) -lm $(SLIB)
.br
.SH RELATED INFORMATION
See the man page ToolsMake.

.SH LOCATION
 readme
