.TH Troubleshooting r "1/3/1995" " " "PETSc"
.SH NAME
Troubleshooting 

What to do when you have trouble linking or compiling PETSc Code!

Most user problems with PETSc come during the compiling and
linking stages. In order to make PORTABLE code one MUST have
portable makefiles. The only way we know how to do this is to
follow the format of the PETSc makefiles very closely. Always start
a new makefile by copying over a makefile from one of the PETSc
examples that already works. LEAVE in any parts of the makefile
that you don't understand, it may be very important.

>> Symptom

alder>make BOPT=g
make: Warning: Can't find `../../bmake/.g': No such file or directory
make: Fatal error in reader: makefile, line 33: Read of include file
`../../bmake/.g' failed

>> Problem

You have not set the variable ARCH to the architecture of your
machine, for instance sun4

>> Cure

Include in your .cshrc file some code to determine the
architecture and set it automatically. For instance
arch >& /dev/null
if ( $status == 0 ) then
setenv ARCH `arch`
else
setenv ARCH unknown
endif

Or
setenv ARCH `tools/bin/tarch`

Or remember to include the ARCH in the command line everytime
you run a make, for instance, make BOPT=g ARCH=sun4 example4

-----------------------------------------------------

>> Symptom

alder>make BOPT=g example1
make: Warning: Can't find `/home/joe/bmake/sun4.g': No such file or
directory
make: Fatal error in reader: makefile, line 33: Read of include file
`/home/joe/bmake/sun4.g' failed

>> Problem

The variable in the makefile ITOOLSDIR does not point to the
tools directory; in this case it points to the directory
/home/joe

>> Cure

Make sure the variable ITOOLSDIR in the makefile points to
the tools directory. Be aware that at many sites, your home
directory may have different names on different machines so
it is usually better to make the path relative, rather than
absolute. That is use ITOOLSDIR = ../../tools.core rather
than /c/cafa/bsmith/tools.core

-----------------------------------------------------

>> Symptom

alder> make BOPT=g ex1
f77 -g -o ex1 ex1.o affine2d.o \\
../../libs/libsg/sun4/domain.a ../../libs/libsg/sun4/Xtools.a
/../libs/libsg/sun4/tools.a  ../../libs/libsg/sun4/liblapack.a
/../libs/libsg/sun4/blas.a ../../libs/libsg/sun4/system.a -lX11  -lm
ld: ex1.o: bad magic number
Compilation failed
*** Error code 4

>> Problem

The file ex1.o was compiled on a different architecture or with a
different compiler.

>> Cure

Remove all .o files and recompile from scatch

----------------------------------------------------------------------

>> Symptom

alder> make BOPT=g ex21
cc  -Dsun4 -g -o ex21 ex21.o affine3d.o ../../libs/libsg/sun4/domain.a
../../libs/libsg/sun4/Xtools.a ../../libs/libsg/sun4/tools.a
../../libs/libsg/sun4/liblapack.a
../../libs/libsg/sun4/blas.a ../../libs/libsg/sun4/system.a
-lX11  -lm
ld: Undefined symbol
_s_cmp
_e_wsfe
_do_f_out
_s_copy
_s_wsFe
_s_stop
Compilation failed
*** Error code 2

>> Problem

You are attempting to link a program which uses Fortran libraries
which are not being found by the linker

>> Cure

Try linking with the Fortran linker (FLINKER) .  This may introduce
additional problems.  If so, Try setting FCLIB in bmake/$ARCH to the
appropriate library, and add $(FCLIB) to the end of the link line.
An example of this use is domain/examples/makefile .

-------------------------------------------------------------------
>> Symptom

alder> make BOPT=g ex21
cc  -Dsun4 -g -o ex21 ex21.o affine3d.o ../../libs/libsg/sun4/domain.a
../../libs/libsg/sun4/Xtools.a ../../libs/libsg/sun4/tools.a
../../libs/libsg/sun4/liblapack.a
../../libs/libsg/sun4/blas.a ../../libs/libsg/sun4/system.a
-lX11 /usr/lang/SC1.0.1/libF77.a  -lm
ld: Undefined symbol
___class_quadruple
Compilation failed
*** Error code 2

>> Problem

You are attempting to link a program which uses Fortran libraries
which are not being found by the linker

>> Cure

See previous Symptom. The FCLIB is set to /usr/lang/SC1.0.1/libF77.a
First, make sure that this is the correct library for your version of
the compiler. If so it is possible that the you should include in
FCLIB /usr/lang/SC1.0.1/libm.a which is sometimes different than the
C compilers libm.a If this doesn't work try linking with the Fortran
linker (FLINKER).  This may introduce additional problems.

--------------------------------------------------------------------

>> Symptom

alder> make BOPT=g example1.sun4
f77 -o example1.sun4 -I../..   -g -DDEBUG_ALL -DDEBUG_TRACEBACK
-DLOGCOMMALL example1.o \\
../../libs/libsg/sun4/tools.a ../../libs/libsg/sun4/tools.a
/../libs/libsg/sun4/tools.a ../../libs/libsg/sun4/system.a   -lm
ld: Undefined symbol
mcount
Compilation failed
*** Error code 2

>> Problem

One of the .o files or libraries has some code in it compiled with
the -pg profiler option (PROFILE=pg) but the linking is being done
without it.

>> Cure

Either recompile the files without the profiling option or
link with the profiling option use make PROFILE=pg ...
alder> make BOPT=g example1.sun4 PROFILE=pg
f77 -pg -o example1.sun4 -I../..   -g -DDEBUG_ALL -DDEBUG_TRACEBACK
-DLOGCOMMALL example1.o \\
../../libs/libsg/sun4/tools.a ../../libs/libsg/sun4/tools.a
/../libs/libsg/sun4/tools.a ../../libs/libsg/sun4/system.a   -lm
alder>

--------------------------------------------------------------------

>> Symptom

alder>make BOPT=g example1.sun4
f77 -o example1.sun4 -I../..   -g -DDEBUG_ALL -DDEBUG_TRACEBACK
-DLOGCOMMALL example1.o \\
../../libs/libsg/sun4/tools.a ../../libs/libsg/sun4/tools.a
/../libs/libsg/sun4/tools.a ../../libs/libsg/sun4/system.a   -lm
ld: Undefined symbol
_p4_get_my_id
_p4_num_total_slaves
Compilation failed
*** Error code 2

>> Problem

The .o files were compiled with the COMM=p4 option but the
linking is done without it. Make sure the link line of libraries
always includes tools$(COMM).a before tools.a

>> Cure

Make sure the linking and compiling are done with exactly the
same options.

--------------------------------------------------------------------

>> Symptom

alder>make BOPT=g example1.sun4 COMM=p4
f77 -o example1.sun4 -I../..  -Dp4 -g -DDEBUG_ALL -DDEBUG_TRACEBACK
-DLOGCOMMALL example1.o \\
../../libs/libsg/sun4/toolsp4.a ../../libs/libsg/sun4/tools.a
/../libs/libsg/sun4/toolsp4.a ../../libs/libsg/sun4/system.a
/c/cafa/bsmith/p4/SUN/lib/libp4.a  -lm
ld: Undefined symbol
___NUMNODES
___MYPROCID
Compilation failed
*** Error code 2

>> Problem

The .o file was compiled with COMM=pvm, while you are linking
with COMM=p4

>> Cure

Make sure the linking and compiling are done with exactly the
same options

--------------------------------------------------------------------

>> Symptom

bamboo>make BOPT=g ex1
cc -Dsolaris -Dsun4  -c  -I../..  -g -DDEBUG_ALL -DDEBUG_TRACEBACK
-DLOGCOMMALL ex1.c
"../../xtools/basex11.h", line 15: cannot find include file: <X11/Xlib.h>
"../../xtools/basex11.h", line 16: cannot find include file: <X11/Xutil.h>
"../../xtools/basex11.h", line 24: syntax error before or at: GC
"../../xtools/basex11.h", line 24: cannot recover from previous errors
cc: acomp failed for ex1.c
*** Error code 2

>> Problem

The standard X11 files are not in the usual place, /usr/include

>> Cure

1) Make sure the makefile has the line
include $(ITOOLSDIR)/bmake/$(ARCH).x11
2) Make sure the file $(ITOOLSDIR)/bmake/$(ARCH).x11
has the correct location of the X11 include files and libraries;
for instance it may have
X11INCLUDEDIR = -I/usr/openwin/include
X11LIB = /usr/openwin/lib/libX11.a
3) Make sure the CFLAGS line in the makefile includes
$(X11INCLUDEDIR)
4) Make sure the list of libaries to be linked in includes
near the end $(X11LIB)

--------------------------------------------------------------------

>> Symptom

f77 -g -o ex1 ex1.o affine2d.o \\
../../libs/libsg/sun4/domain.a ../../libs/libsg/sun4/Xtools.a
/../libs/libsg/sun4/tools.a  ../../libs/libsg/sun4/liblapack.a
/../libs/libsg/sun4/blas.a ../../libs/libsg/sun4/system.a  -lm
ld: Undefined symbol
_XCreateColormap
_XGetWMName
_XSetWMName
_XAllocColor
_XGetImage
_XSetStandardProperties
_XQueryFont
_XGetGeometry


>> Problem

The standard X libraries are not being found

>> Cure

1) Make sure the makefile has the line
include $(ITOOLSDIR)/bmake/$(ARCH).x11
2) Make sure the file $(ITOOLSDIR)/bmake/$(ARCH).x11
has the correct location of the X11 include files and libraries;
for instance it may have
X11INCLUDEDIR = -I/usr/openwin/include
X11LIB = /usr/openwin/lib/libX11.a
3) Make sure the CFLAGS line in the makefile includes
$(X11INCLUDEDIR)
4) Make sure the list of libaries to be linked in includes
near the end $(X11LIB)

--------------------------------------------------------------------

>> Symptom

alder>make BOPT=g example4
f77 -o example4 -g example4.o ../../fort/sun4/fort.a
/../libs/libsg/sun4/tools.a ../../libs/libsg/sun4/Xtools.a
/../libs/libsg/sun4/system.a -lX11      ../../libs/libsg/sun4/blas.a
ld: ../../fort/sun4/fort.a: No such file or directory
Compilation failed
*** Error code 4

>> Problem

For some reason the Fortran interface routines were not compiled.

>> Cure

1) Go to the tools/fort directory
2) Verify that there are a large number of small .c files
3) verify there is a subdirectory with the name $ARCH, for instance
sun4, if it is not there make it.
4) type make BOPT=g and monitor to determine any errors.


>> Symptom

A parallel program thinks that it only has one process, even though more
were specified (e.g., with a.out -np 10 )

>> Problem

The program is using the uni-processor version of the Chameleon
operations.  These are provided to simplify the construction of
sequential versions of parallel programs.

>> Cure

Recompile ALL files, making sure that $(COPT) is included with the
compilation of all C programs, and that tools$(COMM).a comes before
tools.a in the link line.`


.SH LOCATION
 readme
